from __future__ import annotations

import ipaddress
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal

from pydantic import (
    BaseModel,
    Extra,
    Field,
    IPvAnyNetwork,
    ValidationError,
    root_validator,
    validator,
)

import spyctl.spyctl_lib as lib


def valid_object(
    data: Dict, verbose=True, allow_obj_list=True, interactive=False
) -> bool:
    kind = data.get(lib.KIND_FIELD)
    if kind not in KIND_TO_SCHEMA:
        if lib.ITEMS_FIELD not in data:
            lib.err_exit(
                f"Unable to validate {kind!r}, no schema exists for objects of"
                " that type."
            )
        elif not allow_obj_list:
            lib.err_exit("Nested item lists are not allowed.")
        try:
            GuardianObjectListModel(**data)
        except ValidationError as e:
            if verbose:
                if interactive:
                    return str(e)
                else:
                    lib.try_log(str(e), is_warning=True)
            return False
        for item in data[lib.ITEMS_FIELD]:
            if not valid_object(item, allow_obj_list=False):
                return False
        return True
    # Some validations depend on the type of the object in addition to the kind
    tmp_kind = (
        kind,
        data.get(lib.METADATA_FIELD, {}).get(lib.METADATA_TYPE_FIELD),
    )
    if tmp_kind in KIND_TO_SCHEMA:
        kind = tmp_kind
    try:
        KIND_TO_SCHEMA[kind](**data)
    except ValidationError as e:
        if verbose:
            if interactive:
                return str(e)
            else:
                lib.try_log(str(e), is_warning=True)
        return False
    return True


def valid_context(context_data: Dict, verbose=True):
    try:
        ContextsModel(**context_data)
    except ValidationError as e:
        if verbose:
            lib.try_log(str(e), is_warning=True)
        return False
    return True


def handle_show_schema(kind: str) -> str:
    object = KIND_TO_SCHEMA.get(kind)
    return object.schema_json()


def valid_notification_target(tgt_data: Dict, interactive=False):
    try:
        NotificationTargetModel(**tgt_data)
    except ValidationError as e:
        if interactive:
            return str(e)
        else:
            lib.try_log(str(e), is_warning=True)
        return False
    return True


__PROC_IDS = {}

# -----------------------------------------------------------------------------
# Selectors -------------------------------------------------------------------
# -----------------------------------------------------------------------------


class MatchLabelsModel(BaseModel):
    match_labels: Dict[str, str] = Field(alias=lib.MATCH_LABELS_FIELD)

    class Config:
        extra = Extra.forbid


class ContainerSelectorModel(BaseModel):
    image: Optional[str] = Field(alias=lib.IMAGE_FIELD)
    image_id: Optional[str] = Field(alias=lib.IMAGEID_FIELD)
    container_name: Optional[str] = Field(alias=lib.CONTAINER_NAME_FIELD)
    container_id: Optional[str] = Field(alias=lib.CONTAINER_ID_FIELD)

    @root_validator(skip_on_failure=True)
    def ensure_one_field(cls, values: Dict):
        if not any([value for value in values.values()]):
            raise ValueError("")
        return values

    class Config:
        extra = Extra.forbid


class ServiceSelectorModel(BaseModel):
    cgroup: str = Field(alias=lib.CGROUP_FIELD)

    class Config:
        extra = Extra.forbid


class MachineSelectorModel(BaseModel):
    hostname: Optional[Union[str, List[str]]] = Field(alias=lib.HOSTNAME_FIELD)
    machine_uid: Optional[Union[str, List[str]]] = Field(
        alias=lib.MACHINE_UID_FIELD
    )

    @root_validator(pre=True)
    def ensure_one_exists(cls, values: Dict):
        if not any([value for value in values.values()]):
            raise ValueError("")
        return values

    class Config:
        extra = Extra.forbid


class NamespaceSelectorModel(MatchLabelsModel):
    pass


class PodSelectorModel(MatchLabelsModel):
    pass


class TraceSelectorModel(BaseModel):
    trigger_class: Optional[List[str]] = Field(alias=lib.TRIGGER_CLASS_FIELD)
    trigger_ancestor: Optional[List[str]] = Field(
        alias=lib.TRIGGER_ANCESTORS_FIELD
    )

    class Config:
        extra = Extra.forbid


class UserSelectorModel(BaseModel):
    users: Optional[List[str]] = Field(alias=lib.USERS_FIELD)
    interactive_users: Optional[List[str]] = Field(
        alias=lib.INTERACTIVE_USERS_FIELD
    )
    non_interactive_users: Optional[List[str]] = Field(
        alias=lib.NON_INTERACTIVE_USERS_FIELD
    )

    class Config:
        extra = Extra.forbid


class ProcessSelectorModel(BaseModel):
    name: Optional[List[str]] = Field(alias=lib.NAME_FIELD, min_items=1)
    exe: Optional[List[str]] = Field(alias=lib.EXE_FIELD, min_items=1)
    euser: Optional[List[str]] = Field(alias=lib.EUSER_FIELD, min_items=1)
    interactive: Optional[bool] = Field(alias=lib.INTERACTIVE_FIELD)

    @root_validator(skip_on_failure=True)
    def ensure_one_field(cls, values: Dict):
        set_count = 0
        for v in values.values():
            if v is not None:
                set_count += 1
        if set_count == 0:
            raise ValueError("At least one key, value pair expected")
        return values

    class Config:
        extra = Extra.forbid


# -----------------------------------------------------------------------------
# Guardian Models -------------------------------------------------------------
# -----------------------------------------------------------------------------


# This is a reused validator ensuring that the objects have a required selector
def validate_selectors(_, values):
    type = getattr(values["metadata"], "type", "")
    if type == lib.POL_TYPE_CONT:
        s_val = getattr(values["spec"], "container_selector", None)
        if not s_val:
            raise ValueError(
                f"Type is '{lib.POL_TYPE_CONT}' and no "
                f"'{lib.CONT_SELECTOR_FIELD}' found in {lib.SPEC_FIELD}"
            )
    else:
        s_val = getattr(values["spec"], "service_selector", None)
        if not s_val:
            raise ValueError(
                f"Type is '{lib.POL_TYPE_SVC}' and no "
                f"'{lib.SVC_SELECTOR_FIELD}' found in {lib.SPEC_FIELD}"
            )
    return values


class GuardianSelectorsModel(BaseModel):
    container_selector: Optional[ContainerSelectorModel] = Field(
        alias=lib.CONT_SELECTOR_FIELD
    )
    service_selector: Optional[ServiceSelectorModel] = Field(
        alias=lib.SVC_SELECTOR_FIELD
    )
    machine_selector: Optional[MachineSelectorModel] = Field(
        alias=lib.MACHINE_SELECTOR_FIELD
    )
    namespace_selector: Optional[NamespaceSelectorModel] = Field(
        alias=lib.NAMESPACE_SELECTOR_FIELD
    )
    pod_selector: Optional[PodSelectorModel] = Field(
        alias=lib.POD_SELECTOR_FIELD
    )

    class Config:
        extra = Extra.forbid


class ActionSelectorsModel(BaseModel):
    namespace_selector: Optional[NamespaceSelectorModel] = Field(
        alias=lib.NAMESPACE_SELECTOR_FIELD
    )
    pod_selector: Optional[PodSelectorModel] = Field(
        alias=lib.POD_SELECTOR_FIELD
    )
    process_selector: Optional[ProcessSelectorModel] = Field(
        alias=lib.PROCESS_SELECTOR_FIELD
    )

    class Config:
        extra = Extra.forbid


class GuardianSpecOptionsModel(BaseModel):
    disable_processes: Optional[
        Literal[tuple(lib.DISABLE_PROCS_STRINGS)]  # type: ignore
    ] = Field(alias=lib.DISABLE_PROCS_FIELD)
    disable_connections: Optional[
        Literal[tuple(lib.DISABLE_CONNS_STRINGS)]  # type: ignore
    ] = Field(alias=lib.DISABLE_CONNS_FIELD)
    disable_private_conns: Optional[
        Literal[tuple(lib.DISABLE_CONNS_STRINGS)]  # type: ignore
    ] = Field(alias=lib.DISABLE_PR_CONNS_FIELD)
    disable_public_conns: Optional[
        Literal[tuple(lib.DISABLE_CONNS_STRINGS)]  # type: ignore
    ] = Field(alias=lib.DISABLE_PU_CONNS_FIELD)


# Network Models --------------------------------------------------------------


class DnsBlockModel(BaseModel):
    dns_selector: List[str] = Field(alias=lib.DNS_SELECTOR_FIELD)

    class Config:
        extra = Extra.forbid


class CIDRModel(BaseModel):
    cidr: IPvAnyNetwork = Field(alias=lib.CIDR_FIELD)
    except_cidr: Optional[List[IPvAnyNetwork]] = Field(
        alias=lib.EXCEPT_FIELD, max_items=10
    )

    @root_validator(skip_on_failure=True)
    def validate_except_within_cidr(cls, values):
        cidr = values["cidr"]
        try:
            cidr_net = ipaddress.IPv4Network(cidr)
        except ipaddress.AddressValueError:
            cidr_net = ipaddress.IPv6Network(cidr)
        net_type = type(cidr_net)
        if "except_cidr" in values and values["except_cidr"]:
            for e_cidr in values["except_cidr"]:
                try:
                    e_net = ipaddress.IPv4Network(e_cidr)
                except ipaddress.AddressValueError:
                    e_net = ipaddress.IPv6Network(e_cidr)
                if net_type != type(e_net):
                    raise ValueError("Network types are not the same")
                if not cidr_net.supernet_of(e_net):
                    raise ValueError(
                        f"'{e_net}' is not a subnet of '{cidr_net}'"
                    )
        return values

    class Config:
        extra = Extra.forbid


class IpBlockModel(BaseModel):
    ip_block: CIDRModel = Field(alias=lib.IP_BLOCK_FIELD)

    class Config:
        extra = Extra.forbid


class PortsModel(BaseModel):
    port: int = Field(alias=lib.PORT_FIELD, ge=0, le=65535)
    proto: Literal["UDP", "TCP"] = Field(alias=lib.PROTO_FIELD)
    endport: Optional[int] = Field(alias=lib.ENDPORT_FIELD, ge=0, le=66535)

    @root_validator(skip_on_failure=True)
    def endport_ge_port(cls, values):
        endport = values.get("endport")
        if endport is not None and endport < values["port"]:
            raise ValueError(
                f"{lib.ENDPORT_FIELD} must be greater than or equal to"
                f" {lib.PORT_FIELD}"
            )
        return values

    class Config:
        extra = Extra.forbid


class IngressNodeModel(BaseModel):
    from_field: List[Union[DnsBlockModel, IpBlockModel]] = Field(
        alias=lib.FROM_FIELD, min_items=1
    )
    processes: Optional[List[str]] = Field(
        alias=lib.PROCESSES_FIELD, min_items=1
    )
    ports: List[PortsModel] = Field(alias=lib.PORTS_FIELD, min_items=1)

    @validator("processes")
    def validate_proc_ids(cls, v):
        if not v:
            return v
        bad = []
        for proc_id in v:
            if not in_proc_ids(proc_id):
                bad.append(proc_id)
        if bad:
            raise ValueError(f"No process found with id(s) '{bad}'.")
        return v

    class Config:
        smart_union = True
        extra = Extra.forbid


class EgressNodeModel(BaseModel):
    to: List[Union[DnsBlockModel, IpBlockModel]] = Field(
        alias=lib.TO_FIELD, min_items=1
    )
    processes: Optional[List[str]] = Field(
        alias=lib.PROCESSES_FIELD, min_items=1
    )
    ports: List[PortsModel] = Field(alias=lib.PORTS_FIELD, min_items=1)

    @validator("processes")
    def validate_proc_ids(cls, v):
        if not v:
            return v
        bad = []
        for proc_id in v:
            if not in_proc_ids(proc_id):
                bad.append(proc_id)
        if bad:
            raise ValueError(f"No process found with id(s) '{bad}'.")
        return v

    class Config:
        smart_union = True
        extra = Extra.forbid


class NetworkPolicyModel(BaseModel):
    ingress: List[IngressNodeModel] = Field(alias=lib.INGRESS_FIELD)
    egress: List[EgressNodeModel] = Field(alias=lib.EGRESS_FIELD)


class DeviationNetworkPolicyModel(BaseModel):
    ingress: Optional[List[IngressNodeModel]] = Field(alias=lib.INGRESS_FIELD)
    egress: Optional[List[EgressNodeModel]] = Field(alias=lib.EGRESS_FIELD)


# Process Models --------------------------------------------------------------


class SimpleProcessNodeModel(BaseModel):
    name: str = Field(alias=lib.NAME_FIELD)
    exe: List[str] = Field(alias=lib.EXE_FIELD)
    euser: Optional[List[str]] = Field(alias=lib.EUSER_FIELD)


class ProcessNodeModel(SimpleProcessNodeModel):
    id: str = Field(alias=lib.ID_FIELD)
    listening_sockets: Optional[List[PortsModel]] = Field(
        alias=lib.LISTENING_SOCKETS
    )
    children: Optional[List[ProcessNodeModel]] = Field(
        alias=lib.CHILDREN_FIELD
    )

    @validator("id")
    def validate_no_duplicate_ids(cls, v):
        if in_proc_ids(v):
            raise ValueError(f"Duplicate id '{v}' detected.")
        add_proc_id(v)
        return v

    class Config:
        extra = Extra.forbid


class GuardDeviationProcessNodeModel(BaseModel):
    id: str = Field(alias=lib.ID_FIELD)
    children: Optional[List[ProcessNodeModel]] = Field(
        alias=lib.CHILDREN_FIELD
    )

    @validator("id")
    def validate_no_duplicate_ids(cls, v):
        if in_proc_ids(v):
            raise ValueError(f"Duplicate id '{v}' detected.")
        add_proc_id(v)
        return v

    # class Config:
    #     extra = Extra.forbid


class GuardDeviationNodeModel(BaseModel):
    policy_node: GuardDeviationProcessNodeModel = Field(alias="policyNode")

    class Config:
        extra = Extra.forbid


# Actions Models --------------------------------------------------------------


class SharedDefaultActionFieldsModel(BaseModel):
    enabled: Optional[bool] = Field(alias=lib.ENABLED_FIELD)

    class Config:
        extra = Extra.forbid


class SharedActionFieldsModel(ActionSelectorsModel):
    enabled: Optional[bool] = Field(alias=lib.ENABLED_FIELD)

    @root_validator(pre=True)
    def validate_has_selector(cls, fields: Dict):
        count = 0
        values_count = 0
        for key, value in fields.items():
            if key.endswith("Selector"):
                count += 1
            if value:
                values_count += 1
        if count == 0:
            raise ValueError(
                "At least one selector required for non-default actions."
            )
        if values_count != count:
            raise ValueError("All selectors must have values.")
        return fields

    class Config:
        extra = Extra.forbid


class DefaultMakeRedflagModel(SharedDefaultActionFieldsModel):
    content: Optional[str] = Field(alias=lib.FLAG_CONTENT, max_length=350)
    impact: Optional[str] = Field(alias=lib.FLAG_IMPACT, max_length=100)
    severity: Literal[tuple(lib.ALLOWED_SEVERITIES)] = Field(  # type: ignore
        alias=lib.FLAG_SEVERITY
    )

    class Config:
        extra = Extra.forbid


class MakeRedflagModel(SharedActionFieldsModel, DefaultMakeRedflagModel):
    pass


class DefaultMakeOpsflagModel(BaseModel):
    content: Optional[str] = Field(alias=lib.FLAG_CONTENT, max_length=350)
    description: Optional[str] = Field(
        alias=lib.FLAG_DESCRIPTION, max_length=350
    )
    severity: Literal[tuple(lib.ALLOWED_SEVERITIES)] = Field(  # type: ignore
        alias=lib.FLAG_SEVERITY
    )

    class Config:
        extra = Extra.forbid


class MakeOpsflagModel(SharedActionFieldsModel, DefaultMakeOpsflagModel):
    pass


class DefaultWebhookActionModel(BaseModel):
    url_destination: str = Field(
        alias=lib.URL_DESTINATION_FIELD, max_length=2048
    )
    template: Literal[tuple(lib.ALLOWED_TEMPLATES)] = Field(  # type: ignore
        alias=lib.TEMPLATE_FIELD
    )


class WebhookActionModel(SharedActionFieldsModel, DefaultWebhookActionModel):
    pass


class DefaultActionsModel(BaseModel):
    make_redflag: Optional[DefaultMakeRedflagModel] = Field(
        alias=lib.ACTION_MAKE_REDFLAG
    )
    make_opsflag: Optional[DefaultMakeOpsflagModel] = Field(
        alias=lib.ACTION_MAKE_OPSFLAG
    )
    webhook: Optional[DefaultWebhookActionModel] = Field(
        alias=lib.ACTION_WEBHOOK
    )
    agent_kill_pod: Optional[SharedDefaultActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_POD
    )
    agent_kill_proc: Optional[SharedDefaultActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_PROC
    )
    agent_kill_proc_group: Optional[SharedDefaultActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_PROC_GRP
    )

    @root_validator(pre=True)
    def validate_only_one_action(cls, values: Dict):
        actions_count = len(values)
        if actions_count > 1:
            raise ValueError(
                "Detected multiple action definitions in one action. Each"
                " action definition must be a separate entry in the list."
            )
        return values

    class Config:
        extra = Extra.forbid


class ResponseActionsModel(BaseModel):
    make_redflag: Optional[MakeRedflagModel] = Field(
        alias=lib.ACTION_MAKE_REDFLAG
    )
    make_opsflag: Optional[MakeOpsflagModel] = Field(
        alias=lib.ACTION_MAKE_OPSFLAG
    )
    webhook: Optional[WebhookActionModel] = Field(alias=lib.ACTION_WEBHOOK)
    agent_kill_pod: Optional[SharedActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_POD
    )
    agent_kill_proc: Optional[SharedActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_PROC
    )
    agent_kill_proc_group: Optional[SharedActionFieldsModel] = Field(
        alias=lib.ACTION_KILL_PROC_GRP
    )

    @root_validator(pre=True)
    def validate_only_one_action(cls, values: Dict):
        actions_count = len(values)
        if actions_count > 1:
            raise ValueError(
                "Detected multiple action definitions in one action. Each"
                " action definition must be a separate entry in the list."
            )
        return values

    class Config:
        extra = Extra.forbid


class GuardianResponseModel(BaseModel):
    default_field: List[DefaultActionsModel] = Field(
        alias=lib.RESP_DEFAULT_FIELD
    )
    response_field: List[ResponseActionsModel] = Field(
        alias=lib.RESP_ACTIONS_FIELD
    )


# Metadata Models -------------------------------------------------------------


class GuardianMetadataModel(BaseModel):
    name: str = Field(alias=lib.METADATA_NAME_FIELD)
    type: Literal[tuple(lib.GUARDIAN_POL_TYPES)] = Field(  # type: ignore
        alias=lib.METADATA_TYPE_FIELD
    )
    create_time: Optional[Union[int, float, str]] = Field(
        alias=lib.METADATA_CREATE_TIME
    )
    first_timestamp: Optional[Union[int, float, str]] = Field(
        alias=lib.FIRST_TIMESTAMP_FIELD
    )
    latest_timestamp: Optional[Union[int, float, str]] = Field(
        alias=lib.LATEST_TIMESTAMP_FIELD
    )
    uid: Optional[str] = Field(alias=lib.METADATA_UID_FIELD)
    checksum: Optional[str] = Field(alias=lib.METADATA_S_CHECKSUM_FIELD)


class GuardianFingerprintGroupMetadataModel(BaseModel):
    image: Optional[str] = Field(alias=lib.IMAGE_FIELD)
    image_id: Optional[str] = Field(alias=lib.IMAGEID_FIELD)
    cgroup: Optional[str] = Field(alias=lib.CGROUP_FIELD)
    first_timestamp: Optional[Union[int, float]] = Field(
        alias=lib.FIRST_TIMESTAMP_FIELD
    )
    latest_timestamp: Optional[Union[int, float]] = Field(
        alias=lib.LATEST_TIMESTAMP_FIELD
    )


class GuardianDeviationMetadataModel(BaseModel):
    type: str = Field(alias=lib.METADATA_TYPE_FIELD)
    policy_uid: str = Field(alias="policy_uid")


# Spec Models -----------------------------------------------------------------


class GuardianPolicySpecModel(
    GuardianSelectorsModel, GuardianSpecOptionsModel
):
    enabled: Optional[bool] = Field(alias=lib.ENABLED_FIELD)
    mode: Literal[tuple(lib.POL_MODES)] = Field(  # type: ignore
        alias=lib.POL_MODE_FIELD
    )
    process_policy: List[ProcessNodeModel] = Field(alias=lib.PROC_POLICY_FIELD)
    network_policy: NetworkPolicyModel = Field(alias=lib.NET_POLICY_FIELD)
    response: GuardianResponseModel = Field(alias=lib.RESPONSE_FIELD)

    class Config:
        extra = Extra.forbid


class GuardianBaselineSpecModel(
    GuardianSelectorsModel, GuardianSpecOptionsModel
):
    process_policy: List[ProcessNodeModel] = Field(alias=lib.PROC_POLICY_FIELD)
    network_policy: NetworkPolicyModel = Field(alias=lib.NET_POLICY_FIELD)

    class Config:
        extra = Extra.forbid


class GuardianDeviationSpecModel(
    GuardianSelectorsModel, GuardianSpecOptionsModel
):
    process_policy: List[
        Union[ProcessNodeModel, GuardDeviationNodeModel]
    ] = Field(alias=lib.PROC_POLICY_FIELD)
    network_policy: Optional[DeviationNetworkPolicyModel] = Field(
        alias=lib.NET_POLICY_FIELD
    )

    class Config:
        extra = Extra.forbid


# Top-level Models ------------------------------------------------------------


class GuardianFingerprintModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.FPRINT_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: GuardianMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: GuardianBaselineSpecModel = Field(alias=lib.SPEC_FIELD)

    _selector_validator = root_validator(
        allow_reuse=True, skip_on_failure=True
    )(validate_selectors)

    def __init__(self, **data: Any):
        clear_proc_ids()
        super().__init__(**data)
        clear_proc_ids()

    class Config:
        extra = Extra.forbid


class FingerprintGroupDataModel(BaseModel):
    fingerprints: List[GuardianFingerprintModel] = Field(
        alias=lib.FPRINT_GRP_FINGERPRINTS_FIELD
    )
    cont_names: Optional[List[str]] = Field(
        alias=lib.FPRINT_GRP_CONT_NAMES_FIELD
    )
    cont_ids: Optional[List[str]] = Field(alias=lib.FPRINT_GRP_CONT_IDS_FIELD)
    machines: Optional[List[str]] = Field(alias=lib.FPRINT_GRP_MACHINES_FIELD)

    class Config:
        extra = Extra.forbid


class GuardianFingerprintGroupModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.FPRINT_GROUP_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: GuardianFingerprintGroupMetadataModel = Field(
        alias=lib.METADATA_FIELD
    )
    data: FingerprintGroupDataModel

    class Config:
        extra = Extra.ignore


class GuardianDeviationModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.DEVIATION_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: GuardianDeviationMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: GuardianDeviationSpecModel = Field(alias=lib.SPEC_FIELD)

    def __init__(self, **data: Any):
        clear_proc_ids()
        super().__init__(**data)
        clear_proc_ids()


class GuardianBaselineModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.BASELINE_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: GuardianMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: GuardianBaselineSpecModel = Field(alias=lib.SPEC_FIELD)

    _selector_validator = root_validator(
        allow_reuse=True, skip_on_failure=True
    )(validate_selectors)

    def __init__(self, **data: Any):
        clear_proc_ids()
        super().__init__(**data)
        clear_proc_ids()

    class Config:
        extra = Extra.ignore


class GuardianPolicyModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.POL_KIND] = Field(alias=lib.KIND_FIELD)  # type: ignore
    metadata: GuardianMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: GuardianPolicySpecModel = Field(alias=lib.SPEC_FIELD)

    _selector_validator = root_validator(
        allow_reuse=True, skip_on_failure=True
    )(validate_selectors)

    def __init__(self, **data: Any):
        clear_proc_ids()
        super().__init__(**data)
        clear_proc_ids()

    class Config:
        extra = Extra.forbid


class GuardianObjectModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: str = Field(alias=lib.KIND_FIELD)
    metadata: Dict[str, str] = Field(alias=lib.METADATA_FIELD)
    spec: Dict = Field(alias=lib.SPEC_FIELD)

    class Config:
        extra = Extra.ignore


class GuardianObjectListModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    items: List[GuardianObjectModel] = Field(alias=lib.ITEMS_FIELD)


# -----------------------------------------------------------------------------
# Notification Models ---------------------------------------------------------
# -----------------------------------------------------------------------------


class DestinationSlackModel(BaseModel):
    url: str = Field(alias=lib.DST_SLACK_URL)

    @validator("url")
    def valid_url(cls, url):
        if not lib.is_valid_slack_url(url):
            raise ValueError(
                "Invalid url format. Example: https://hooks.slack.com/services/xxxxxxxxxxx/xxxxxxxxxxx/xxxxxxxxxxxxxxxxxxxxxxxx"
            )
        return url

    class Config:
        extra = Extra.forbid


class DestinationWebhookModel(BaseModel):
    url: str = Field(alias=lib.DST_WEBHOOK_URL)
    no_tls_validation: bool = Field(alias=lib.DST_WEBHOOK_TLS_VAL)

    @validator("url")
    def valid_url(cls, url):
        if not lib.is_valid_url(url):
            raise ValueError(
                "Invalid url format. Example: https://my.url.example"
            )
        return url

    class Config:
        extra = Extra.forbid


class DestinationSNSModel(BaseModel):
    cross_account_iam_role: Optional[str] = Field(
        alias=lib.DST_SNS_CROSS_ACCOUNT_ROLE
    )
    sns_topic_arn: str = Field(alias=lib.DST_SNS_TOPIC_ARN)

    class Config:
        extra = Extra.forbid


class AllDestinationsModel(BaseModel):
    org_uid: Optional[str] = Field(alias=lib.DST_TYPE_ORG)
    emails: Optional[List[str]] = Field(alias=lib.DST_TYPE_EMAIL)
    users: Optional[List[str]] = Field(alias=lib.DST_TYPE_USERS)
    slack: Optional[DestinationSlackModel] = Field(alias=lib.DST_TYPE_SLACK)
    webhook: Optional[DestinationWebhookModel] = Field(
        alias=lib.DST_TYPE_WEBHOOK
    )
    sns: Optional[DestinationSNSModel] = Field(alias=lib.DST_TYPE_SNS)

    @validator("emails", each_item=True)
    def valid_emails(cls, email):
        if not lib.is_valid_email(email):
            raise ValueError("Email format is invalid.")
        return email

    @root_validator(pre=True)
    def ensure_not_none(cls, values):
        for key, value in values.items():
            if value is None:
                raise ValueError(f"Field '{key}' cannot be None")
        return values

    @root_validator(pre=True)
    def one_destination(cls, values: Dict):
        count = 0
        for dst_type in lib.DST_TYPES:
            if dst_type in values:
                count += 1
        if count == 0:
            raise ValueError(
                f"One destination type is required. {lib.DST_TYPES}"
            )
        elif count > 1:
            raise ValueError("Only one destination type is allowed.")
        return values

    class Config:
        extra = Extra.forbid


class NotificationTargetModel(AllDestinationsModel):
    data: Optional[Dict] = Field(alias=lib.DST_DATA)
    description: Optional[str] = Field(
        alias=lib.DST_DESCRIPTION, max_length=128
    )

    class Config:
        extra = Extra.forbid


class NotifTgtMetadataModel(BaseModel):
    name: str = Field(alias=lib.METADATA_NAME_FIELD)
    uid: str = Field(alias=lib.METADATA_UID_FIELD)
    create_time: Optional[Union[float, int]] = Field(
        alias=lib.METADATA_CREATE_TIME
    )
    update_time: Optional[Union[float, int]] = Field(
        alias=lib.NOTIF_LAST_UPDATED
    )


class NotifTgtSpecModel(AllDestinationsModel):
    pass


class NotificationTgtResourceModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.TARGET_KIND] = Field(alias=lib.KIND_FIELD)  # type: ignore
    metadata: NotifTgtMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: NotifTgtSpecModel = Field(alias=lib.SPEC_FIELD)

    class Config:
        extra = Extra.forbid


class NotifAnaConfigMetadataModel(BaseModel):
    name: str = Field(alias=lib.METADATA_NAME_FIELD)
    uid: str = Field(alias=lib.METADATA_UID_FIELD)
    notif_type: Literal[lib.NOTIF_TYPE_OBJECT] = Field(  # type: ignore
        lib.METADATA_TYPE_FIELD
    )
    create_time: Optional[Union[float, int]] = Field(
        alias=lib.METADATA_CREATE_TIME
    )
    update_time: Optional[Union[float, int]] = Field(
        alias=lib.NOTIF_LAST_UPDATED
    )


class NotifAnaConfigMetricsMetadataModel(NotifAnaConfigMetadataModel):
    notif_type: Literal[lib.NOTIF_TYPE_METRICS] = Field(  # type: ignore
        lib.METADATA_TYPE_FIELD
    )


class NotifAdditionalFieldsModel(BaseModel):
    details: Optional[Dict[str, str]]
    linkback_text: Optional[str]
    linkback_url: Optional[str]
    slack_icon: Optional[str]

    @root_validator(pre=True)
    def ensure_not_none(cls, values):
        for key, value in values.items():
            if value is None:
                raise ValueError(f"Field '{key}' cannot be None")
        return values


class NotifAdvancedCooldown(BaseModel):
    by_field: Union[str, List[str]] = Field(
        alias=lib.NOTIF_COOLDOWN_BY_FIELD_FIELD, min_items=1, min_length=1
    )
    for_seconds: int = Field(alias=lib.NOTIF_COOLDOWN_SECONDS_FIELD)


class NotifAnaConfigSpecModel(BaseModel):
    enabled: Optional[bool] = Field(alias=lib.ENABLED_FIELD)
    condition: str = Field(alias=lib.NOTIF_CONDITION_FIELD)
    message: str = Field(alias=lib.NOTIF_MESSAGE_FIELD)
    target: Union[str, List[str]] = Field(
        alias=lib.NOTIF_TARGET_FIELD, min_items=1, min_length=1
    )
    schema_type: str = Field(alias=lib.NOTIF_DEFAULT_SCHEMA)
    sub_schema: Optional[str] = Field(alias=lib.NOTIF_SUB_SCHEMA)
    title: str = Field(alias=lib.NOTIF_TITLE_FIELD)
    additional_fields: Dict = Field(alias=lib.NOTIF_ADDITIONAL_FIELDS)
    template: str = Field(alias=lib.NOTIF_TEMPLATE_FIELD)
    cooldown: Optional[Union[int, NotifAdvancedCooldown]] = Field(
        alias=lib.NOTIF_COOLDOWN_FIELD
    )

    @root_validator
    def validate_condition(cls, values):
        import spyctl.config.configs as cfg
        import spyctl.api as api

        ctx = cfg.get_current_context()
        error = api.validate_search_query(
            *ctx.get_api_data(), values["schema_type"], values["condition"]
        )
        if error and False:
            raise ValueError(error)
        return values

    class Config:
        extra = Extra.forbid


class NotifAnaConfigMetricsSpecModel(NotifAnaConfigSpecModel):
    for_duration: Optional[int] = Field(alias=lib.NOTIF_FOR_DURATION_FIELD)
    schema_type: Literal[lib.EVENT_METRICS_PREFIX] = Field(  # type: ignore
        alias=lib.NOTIF_DEFAULT_SCHEMA
    )

    class Config:
        extra = Extra.forbid


class NotificationConfigModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.NOTIFICATION_KIND] = Field(alias=lib.KIND_FIELD)  # type: ignore
    metadata: NotifAnaConfigMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: NotifAnaConfigSpecModel = Field(alias=lib.SPEC_FIELD)

    class Config:
        extra = Extra.forbid


class NotificationConfigMetricsModel(NotificationConfigModel):
    metadata: NotifAnaConfigMetricsMetadataModel = Field(
        alias=lib.METADATA_FIELD
    )
    spec: NotifAnaConfigMetricsSpecModel = Field(alias=lib.SPEC_FIELD)

    class Config:
        extra = Extra.forbid


class NotificationRouteDataModel(BaseModel):
    analytics_settings: Optional[NotificationConfigModel]


class NotificationRouteModel(BaseModel):
    targets: Optional[List[str]] = Field(alias=lib.ROUTE_TARGETS)
    destination: Optional[NotificationConfigModel] = Field(
        alias=lib.ROUTE_DESTINATION
    )
    data: Optional[Dict] = Field(alias=lib.ROUTE_DATA)
    description: Optional[str] = Field(alias=lib.ROUTE_DESCRIPTION)
    expr: Optional[Dict] = Field(alias=lib.ROUTE_EXPR)

    class Config:
        extra = Extra.forbid


class NotificationPolicyModel(BaseModel):
    targets: Optional[Dict[str, NotificationTargetModel]] = Field(
        alias=lib.TARGETS_FIELD
    )
    routes: Optional[List[NotificationRouteModel]] = Field(
        alias=lib.ROUTES_FIELD
    )

    @validator("targets")
    def validate_target_name(cls, v):
        for name in v:
            if len(name) > 64:
                raise ValueError(
                    "Target name must be less than 64 characters."
                )
            if not lib.is_valid_tgt_name(name):
                raise ValueError(
                    "Target name must be only letters, numbers, and valid"
                    f" symbols {lib.TGT_NAME_VALID_SYMBOLS}."
                )
        return v

    class Config:
        extra = Extra.forbid


# -----------------------------------------------------------------------------
# Suppression Models ----------------------------------------------------------
# -----------------------------------------------------------------------------


class SuppressionPolicySelectorsModel(BaseModel):
    trace_selector: Optional[TraceSelectorModel] = Field(
        alias=lib.TRACE_SELECTOR_FIELD
    )
    user_selector: Optional[UserSelectorModel] = Field(
        alias=lib.USER_SELECTOR_FIELD
    )

    @root_validator
    def ensure_one_field(cls, values: Dict):
        if not any(
            value
            for field, value in values.items()
            if field.endswith("selector")
        ):
            # TODO fill out error
            raise ValueError("")
        return values

    class Config:
        extra = Extra.forbid


# Metadata Models -------------------------------------------------------------


class SuppressionPolicyMetadataModel(BaseModel):
    name: str = Field(alias=lib.METADATA_NAME_FIELD)
    type: Literal[tuple(lib.SUPPRESSION_POL_TYPES)] = Field(  # type: ignore
        alias=lib.METADATA_TYPE_FIELD
    )
    create_time: Optional[Union[int, float, str]] = Field(
        alias=lib.METADATA_CREATE_TIME
    )
    first_timestamp: Optional[Union[int, float]] = Field(
        alias=lib.FIRST_TIMESTAMP_FIELD
    )
    latest_timestamp: Optional[Union[int, float]] = Field(
        alias=lib.LATEST_TIMESTAMP_FIELD
    )
    uid: Optional[str] = Field(alias=lib.METADATA_UID_FIELD)
    checksum: Optional[str] = Field(alias=lib.METADATA_S_CHECKSUM_FIELD)


# Spec Models -----------------------------------------------------------------


class AllowedFlagsModel(BaseModel):
    class_field: str = Field(alias=lib.FLAG_CLASS)


class SuppressionPolicySpecModel(SuppressionPolicySelectorsModel):
    enabled: Optional[bool] = Field(alias=lib.ENABLED_FIELD)
    mode: Literal[tuple(lib.POL_MODES)] = Field(  # type: ignore
        alias=lib.POL_MODE_FIELD
    )
    allowed_flags: List[AllowedFlagsModel] = Field(
        alias=lib.ALLOWED_FLAGS_FIELD
    )

    class Config:
        extra = Extra.forbid


# Top-level Models ------------------------------------------------------------


class SuppressionPolicyModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.POL_KIND] = Field(alias=lib.KIND_FIELD)  # type: ignore
    metadata: SuppressionPolicyMetadataModel = Field(alias=lib.METADATA_FIELD)
    spec: SuppressionPolicySpecModel = Field(alias=lib.SPEC_FIELD)

    class Config:
        extra = Extra.forbid


# -----------------------------------------------------------------------------
# Config Models ----------------------------------------------------------
# -----------------------------------------------------------------------------


class ContextModel(BaseModel):
    org: str = Field(alias=lib.ORG_FIELD)
    cgroups: Optional[Union[str, List[str]]] = Field(alias=lib.CGROUP_FIELD)
    cluster: Optional[Union[str, List[str]]] = Field(alias=lib.CLUSTER_FIELD)
    container_ids: Optional[Union[str, List[str]]] = Field(
        alias=lib.CONTAINER_ID_FIELD
    )
    container_names: Optional[Union[str, List[str]]] = Field(
        alias=lib.CONTAINER_NAME_FIELD
    )
    image_ids: Optional[Union[str, List[str]]] = Field(alias=lib.IMAGEID_FIELD)
    images: Optional[Union[str, List[str]]] = Field(alias=lib.IMAGE_FIELD)
    machines: Optional[Union[str, List[str]]] = Field(alias=lib.MACHINES_FIELD)
    namespace: Optional[Union[str, List[str]]] = Field(
        alias=lib.NAMESPACE_FIELD
    )
    pods: Optional[Union[str, List[str]]] = Field(alias=lib.POD_FIELD)

    class Config:
        extra = Extra.forbid


# Metadata Models -------------------------------------------------------------


class SecretMetadataModel(BaseModel):
    name: str = Field(alias=lib.METADATA_NAME_FIELD)
    create_time: Optional[Union[float, int]] = Field(
        alias=lib.METADATA_CREATE_TIME
    )


# Top-level Models ------------------------------------------------------------


class SecretModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.SECRET_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: SecretMetadataModel = Field(alias=lib.METADATA_FIELD)
    data: Optional[Dict[str, str]] = Field(alias=lib.DATA_FIELD)
    string_data: Optional[Dict[str, str]] = Field(alias=lib.STRING_DATA_FIELD)

    class Config:
        extra = Extra.forbid


class ContextsModel(BaseModel):
    context_name: str = Field(alias=lib.CONTEXT_NAME_FIELD)
    secret: str = Field(alias=lib.SECRET_FIELD)
    context: ContextModel = Field(alias=lib.CONTEXT_FIELD)

    class Config:
        extra = Extra.forbid


class ConfigModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.CONFIG_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    contexts: List[ContextsModel] = Field(alias=lib.CONTEXTS_FIELD)
    current_context: str = Field(alias=lib.CURR_CONTEXT_FIELD)

    class Config:
        extra = Extra.forbid


# -----------------------------------------------------------------------------
# Misc Models -----------------------------------------------------------------
# -----------------------------------------------------------------------------


class UidListMetadataModel(BaseModel):
    start_time: Union[int, float] = Field(alias=lib.METADATA_START_TIME_FIELD)
    end_time: Union[int, float] = Field(alias=lib.METADATA_END_TIME_FIELD)

    @root_validator
    def valid_end_time(cls, values: Dict):
        start_time = values["start_time"]
        end_time = values["end_time"]
        if end_time <= start_time:
            raise ValueError(
                f"'{lib.METADATA_END_TIME_FIELD}' must be greater than"
                f" '{lib.METADATA_START_TIME_FIELD}'"
            )
        return values


class UidListDataModel(BaseModel):
    uids: List[str] = Field(alias=lib.UIDS_FIELD)

    class Config:
        extra = Extra.forbid


class UidListModel(BaseModel):
    api_version: str = Field(alias=lib.API_FIELD)
    kind: Literal[lib.UID_LIST_KIND] = Field(  # type: ignore
        alias=lib.KIND_FIELD
    )
    metadata: UidListMetadataModel = Field(alias=lib.METADATA_FIELD)
    data: UidListDataModel = Field(alias=lib.DATA_FIELD)

    class Config:
        extra = Extra.forbid


class SpyderbatObject(BaseModel):
    api_version: Literal[lib.API_VERSION] = Field(  # type: ignore
        alias=lib.API_FIELD
    )
    kind: str = Field(alias=lib.KIND_FIELD)

    @validator("kind")
    def valid_kind(cls, v):
        if v not in KIND_TO_SCHEMA:
            raise ValueError(f"Kind '{v}' not in {list(KIND_TO_SCHEMA)}")
        return v

    class Config:
        extra = Extra.allow


KIND_TO_SCHEMA: Dict[str, BaseModel] = {
    lib.BASELINE_KIND: GuardianBaselineModel,
    lib.CONFIG_KIND: ConfigModel,
    lib.FPRINT_GROUP_KIND: GuardianFingerprintGroupModel,
    lib.FPRINT_KIND: GuardianFingerprintModel,
    lib.POL_KIND: GuardianPolicyModel,
    (lib.POL_KIND, lib.POL_TYPE_TRACE): SuppressionPolicyModel,
    lib.SECRET_KIND: SecretModel,
    lib.UID_LIST_KIND: UidListModel,
    lib.DEVIATION_KIND: GuardianDeviationModel,
    lib.NOTIFICATION_KIND: NotificationConfigModel,
    (
        lib.NOTIFICATION_KIND,
        lib.NOTIF_TYPE_METRICS,
    ): NotificationConfigMetricsModel,
    lib.TARGET_KIND: NotificationTgtResourceModel,
}


def clear_proc_ids():
    global __PROC_IDS
    __PROC_IDS.clear()


def in_proc_ids(id: str) -> bool:
    return id in __PROC_IDS


def add_proc_id(id: str):
    global __PROC_IDS
    __PROC_IDS[id] = True



from typing import Dict, List, Optional, IO

import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.resources.baselines as b
import spyctl.resources.policies as p
import spyctl.resources.suppression_policies as sp
import spyctl.resources.notification_targets as nt
import spyctl.resources.notifications_configs as nc
import spyctl.search as search
import spyctl.spyctl_lib as lib
import spyctl.api as api


def handle_create_baseline(
    filename: str,
    output: str,
    name: str,
    disable_procs: str,
    disable_conns: str,
):
    resrc_data = lib.load_resource_file(filename)
    baseline = b.create_baseline(
        resrc_data,
        name,
        disable_procs=disable_procs,
        disable_conns=disable_conns,
    )
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(baseline, output)


def handle_create_guardian_policy(
    file: IO,
    output: str,
    name: str,
    mode: str,
    disable_procs: str,
    disable_conns: str,
    do_api=False,
):
    if do_api:
        ctx = cfg.get_current_context()
        resrc_data = lib.load_file_for_api_test(file)
        policy = api.api_create_guardian_policy(
            *ctx.get_api_data(), name, mode, resrc_data
        )
        cli.show(policy, lib.OUTPUT_RAW)
    else:
        policy = create_guardian_policy_from_file(
            file, name, mode, disable_procs, disable_conns
        )
        if output == lib.OUTPUT_DEFAULT:
            output = lib.OUTPUT_YAML
        cli.show(policy, output)


def create_guardian_policy_from_file(
    file: IO, name: str, mode: str, disable_procs: str, disable_conns: str
):
    resrc_data = lib.load_resource_file(file)
    policy = p.create_policy(
        resrc_data,
        name=name,
        mode=mode,
        disable_procs=disable_procs,
        disable_conns=disable_conns,
    )
    return policy


def create_guardian_policy_from_json(
    name: str, mode: str, input_objects: List[Dict], ctx: cfg.Context
):
    policy = p.create_policy(input_objects, mode=mode, name=name, ctx=ctx)
    return policy


def handle_create_suppression_policy(
    type: str,
    id: Optional[str],
    include_users: bool,
    output: str,
    mode: str,
    name: str = None,
    do_api: bool = False,
    **selectors,
):
    if type == lib.POL_TYPE_TRACE:
        handle_create_trace_suppression_policy(
            id, include_users, output, mode, name, do_api, **selectors
        )


def handle_create_trace_suppression_policy(
    id,
    include_users,
    output,
    mode: str,
    name: str = None,
    do_api: bool = False,
    **selectors,
):
    if do_api:
        ctx = cfg.get_current_context()
        policy = api.api_create_suppression_policy(
            *ctx.get_api_data(),
            name,
            lib.POL_TYPE_TRACE,
            include_users,
            id,
            **selectors,
        )
        cli.show(policy, lib.OUTPUT_RAW)
    else:
        pol = create_trace_suppression_policy(
            id, include_users, mode, name, **selectors
        )
        if output == lib.OUTPUT_DEFAULT:
            output = lib.OUTPUT_YAML
        cli.show(pol.as_dict(), output)


def create_trace_suppression_policy(
    id,
    include_users,
    mode,
    name: str = None,
    ctx: cfg.Context = None,
    **selectors,
) -> sp.TraceSuppressionPolicy:
    if id:
        trace = search.search_for_trace_by_uid(id, ctx)
        if not trace:
            exit(1)
        summary_uid = trace.get(lib.TRACE_SUMMARY_FIELD)
        if not summary_uid:
            cli.err_exit(f"Unable to find a Trace Summary for Trace {id}")
        t_sum = search.search_for_trace_summary_by_uid(summary_uid, ctx)
        if not t_sum:
            exit(1)
        pol = sp.build_trace_suppression_policy(
            t_sum, include_users, mode=mode, name=name, **selectors
        )
    else:
        pol = sp.build_trace_suppression_policy(
            mode=mode, name=name, **selectors
        )
    return pol


def handle_create_flag_suppression_policy(
    id, include_users, output, **selectors
):
    pass


def handle_create_notif_tgt(name, type, output):
    target = nt.create_target(name, type)
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(target, output)


def handle_create_notif_config(
    name: str, target: str, template: str, output: str
):
    config = nc.create_config(name, target, template)
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(config, output)



import json
from typing import Dict, List

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.resources.policies as p
import spyctl.resources.suppression_policies as sp
import spyctl.spyctl_lib as lib
import spyctl.commands.merge as m
import spyctl.resources.notification_targets as nt
import spyctl.resources.notifications_configs as nc


def handle_apply(filename):
    resrc_data = lib.load_resource_file(filename)
    kind = resrc_data.get(lib.KIND_FIELD)
    if kind == lib.POL_KIND:
        type = resrc_data[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
        if type in lib.SUPPRESSION_POL_TYPES:
            handle_apply_suppression_policy(resrc_data)
        elif type in lib.GUARDIAN_POL_TYPES:
            handle_apply_policy(resrc_data)
        else:
            cli.err_exit(f"Unrecognized policy type '{type}'.")
    elif kind == lib.NOTIFICATION_KIND:
        handle_apply_notification_config(resrc_data)
    elif kind == lib.TARGET_KIND:
        handle_apply_notification_target(resrc_data)
    else:
        cli.err_exit(f"The 'apply' command is not supported for {kind}")


def handle_apply_policy(policy: Dict):
    ctx = cfg.get_current_context()
    policy = p.Policy(policy)
    uid, api_data = p.get_data_for_api_call(policy)
    if uid:
        resp = api.put_policy_update(*ctx.get_api_data(), uid, api_data)
        if resp.status_code == 200:
            cli.try_log(f"Successfully updated policy {uid}")
    else:
        resp = api.post_new_policy(*ctx.get_api_data(), api_data)
        if resp and resp.text:
            uid = json.loads(resp.text).get("uid", "")
            cli.try_log(f"Successfully applied new policy with uid: {uid}")


def handle_apply_suppression_policy(policy: Dict):
    ctx = cfg.get_current_context()
    matching_policies = check_suppression_policy_selector_hash(policy)
    if matching_policies:
        policy = handle_matching_policies(policy, matching_policies)
    else:
        policy = sp.TraceSuppressionPolicy(policy)
    uid, api_data = sp.get_data_for_api_call(policy)
    if uid:
        resp = api.put_policy_update(*ctx.get_api_data(), uid, api_data)
        if resp.status_code == 200:
            cli.try_log(f"Successfully updated suppression policy {uid}")
    else:
        resp = api.post_new_policy(*ctx.get_api_data(), api_data)
        if resp and resp.text:
            uid = json.loads(resp.text).get("uid", "")
            cli.try_log(
                f"Successfully applied new suppression policy with uid: {uid}"
            )


def handle_matching_policies(policy: Dict, matching_policies: Dict[str, Dict]):
    uid = policy[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    if uid:
        return sp.TraceSuppressionPolicy(policy)
    query = (
        "There already exists a policy matching this scope. Would you like"
        " to merge this policy into the existing one?"
    )
    if not cli.query_yes_no(query):
        return sp.TraceSuppressionPolicy(policy)
    ret_pol = policy
    for uid, m_policy in matching_policies.items():
        merged = m.merge_resource(ret_pol, "", m_policy)
        if merged:
            ret_pol = merged.get_obj_data()
    ret_pol[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = uid
    return sp.TraceSuppressionPolicy(ret_pol)


def handle_apply_notification_target(notif_target: Dict):
    ctx = cfg.get_current_context()
    target = nt.Target(target_resource=notif_target)
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    targets: Dict = notif_pol.get(lib.TARGETS_FIELD, {})
    old_tgt = None
    for tgt_name, tgt_data in targets.items():
        tgt_id = tgt_data.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not tgt_id:
            continue
        if tgt_id == target.id:
            old_tgt = {tgt_name: tgt_data}
            break
        if tgt_name == target.name:
            cli.err_exit("Target names must be unique!")
    if old_tgt:
        tgt_name = next(iter(old_tgt))
        targets.pop(tgt_name)
    target.set_last_update_time()
    new_tgt = target.as_target()
    targets.update(**new_tgt)
    notif_pol[lib.TARGETS_FIELD] = targets
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    if old_tgt:
        cli.try_log(f"Successfully updated Notification Target '{target.id}'")
    else:
        cli.try_log(f"Successfully applied Notification Target '{target.id}'")


def handle_apply_notification_config(notif_config: Dict):
    ctx = cfg.get_current_context()
    config = nc.NotificationConfig(config_resource=notif_config)
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    routes: List[Dict] = notif_pol.get(lib.ROUTES_FIELD, [])
    old_route_index = None
    for i, route in enumerate(routes):
        route_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not route_id:
            continue
        if route_id == config.id:
            old_route_index = i
    if old_route_index is not None:
        routes.pop(i)
    config.set_last_updated()
    new_route = config.route
    routes.append(new_route)
    notif_pol[lib.ROUTES_FIELD] = routes
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    if old_route_index:
        cli.try_log(f"Successfully updated Notification Config '{config.id}'")
    else:
        cli.try_log(f"Successfully applied Notification Config '{config.id}'")


def check_suppression_policy_selector_hash(policy: Dict) -> Dict[str, Dict]:
    ctx = cfg.get_current_context()
    selector_hash = policy[lib.METADATA_FIELD].get(
        lib.METADATA_S_CHECKSUM_FIELD
    )
    if not selector_hash:
        return {}
    type = policy[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    params = {
        lib.METADATA_TYPE_FIELD: type,
        lib.API_HAS_TAGS_FIELD: [f"SELECTOR_HASH:{selector_hash}"],
    }
    resp_policies = api.get_policies(*ctx.get_api_data(), params)
    rv = {}
    for r_pol in resp_policies:
        uid = r_pol[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
        rv[uid] = r_pol
    return rv



from dataclasses import dataclass
from typing import Dict

import spyctl.schemas_v2 as schemas

import app.app_lib as app_lib

# ------------------------------------------------------------------------------
# Validate Spyderbat Object
# ------------------------------------------------------------------------------


@dataclass
class ValidateInput:
    object: Dict


@dataclass
class ValidateOutput:
    invalid_message: str


def validate(i: ValidateInput) -> ValidateOutput:
    if not schemas.valid_object(i.object):
        invalid_msg = app_lib.flush_spyctl_log_messages()
    else:
        invalid_msg = ""
    return ValidateOutput(invalid_message=invalid_msg)



TEMPLATES = [
    {
        "display_name": "Agent Offline",
        "id": "nc_tmpl:000000001",
        "description": "Send a notification when Spyderbat detects it is no longer receiving data from an Agent (Nano Agent or Clustermonitor, ephemeral or not).",
        "type": "agent_health",
        "config": {
            "schema_type": "event_opsflag",
            "sub_schema": "agent_offline",
            "condition": "",
            "title": "Spyderbat Agent Detected Offline",
            "message": "{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Hostname": "{{ hostname }}",
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Cluster": "{{ __cluster__ }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":red_circle:",
            },
        },
    },
    {
        "display_name": "Agent Back Online",
        "id": "nc_tmpl:000000002",
        "description": "Send a notification when Spyderbat detects that an Agent has come back online and is sending data.",
        "type": "agent_health",
        "config": {
            "schema_type": "event_opsflag",
            "sub_schema": "agent_online",
            "condition": "",
            "title": "Spyderbat Agent Back Online",
            "message": "{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Hostname": "{{ hostname }}",
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Cluster": "{{ __cluster__ }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":large_green_circle:",
            },
        },
    },
    {
        "display_name": "Agent CPU Usage Over Threshold",
        "id": "nc_tmpl:000000003",
        "description": "Send a notification when an Agent's total CPU usage is over 4% for 2 minutes. (30 minute cooldown)",
        "type": "agent_health",
        "config": {
            "schema_type": "event_metric",
            "sub_schema": "agent",
            "condition": "cpu_1min_P.agent > 0.04",
            "for_duration": 120,
            "cooldown": 1800,
            "title": "Spyderbat Agent CPU Over Threshold For 2 Minutes",
            "message": "{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Hostname": "{{ hostname }}",
                    "Cluster": "{{ __cluster__ }}",
                    "CPU Used (%)": "{{ __percent__ | cpu_1min_P.agent }} (Threshold: 4%)",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":large_yellow_circle:",
            },
        },
    },
    {
        "display_name": "Agent Memory Over Threshold",
        "id": "nc_tmpl:000000004",
        "description": "Send a notification when an Agent's memory usage is over 3.5GB for 2 minutes. (30 minute cooldown)",
        "type": "agent_health",
        "config": {
            "schema_type": "event_metric",
            "sub_schema": "agent",
            "condition": "mem_1min_B.agent > 3758096384",
            "for_duration": 120,
            "cooldown": 1800,
            "title": "Spyderbat Agent Memory Usage Over Threshold For 2 Minutes",
            "message": "{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Hostname": "{{ hostname }}",
                    "Cluster": "{{ __cluster__ }}",
                    "Memory Used (%)": "{{ __percent__ | mem_1min_P.agent }}",
                    "Memory Used (bytes)": "{{ mem_1min_B.agent }}B (Threshold: 3.5GB)",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":large_yellow_circle:",
            },
        },
    },
    {
        "display_name": "Agent Bandwidth Over Threshold",
        "id": "nc_tmpl:000000005",
        "description": "Send a notification when an Agent's bandwidth usage is over 125 KBps for 2 minutes. (30 minute cooldown)",
        "type": "agent_health",
        "config": {
            "schema_type": "event_metric",
            "sub_schema": "agent",
            "condition": "bandwidth_1min_Bps > 125000",
            "for_duration": 120,
            "cooldown": 1800,
            "title": "Spyderbat Agent Bandwidth Usage Over Threshold For 2 Minutes",
            "message": "{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Hostname": "{{ hostname }}",
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Cluster": "{{ __cluster__ }}",
                    "Bandwidth Used (Bps)": "{{ bandwidth_1min_Bps }} Bps (Threshold: 125,000 Bps)",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":large_yellow_circle:",
            },
        },
    },
    {
        "display_name": "Bat Offline",
        "id": "nc_tmpl:000000006",
        "description": "Send a notification when a Bat goes offline.",
        "type": "agent_health",
        "config": {
            "schema_type": "event_opsflag",
            "sub_schema": "bat_offline",
            "condition": "",
            "title": "Spyderbat Bat Offline",
            "message": "{{ description }}\n\n{{ __origin__ }}",
            "cooldown": 900,
            "additional_fields": {
                "details": {
                    "Hostname": "{{ hostname }}",
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Bat": "{{ bat_name }}",
                    "Severity": "{{ severity }}",
                    "Cluster": "{{ __cluster__ }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":red_circle:",
            },
        },
    },
    {
        "display_name": "Bat Online",
        "id": "nc_tmpl:000000007",
        "description": "Send a notification when a Bat comes back online.",
        "type": "agent_health",
        "config": {
            "schema_type": "event_opsflag",
            "sub_schema": "bat_online",
            "condition": "",
            "title": "Spyderbat Bat Back Online",
            "message": "{{ description }}\n\n{{ __origin__ }}",
            "cooldown": 900,
            "additional_fields": {
                "details": {
                    "Hostname": "{{ hostname }}",
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Bat": "{{ bat_name }}",
                    "Severity": "{{ severity }}",
                    "Cluster": "{{ __cluster__ }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":large_green_circle:",
            },
        },
    },
    {
        "display_name": "SSH Login Detection",
        "id": "nc_tmpl:00000008",
        "description": "Send a notification when Spyderbat detects an interactive SSH login.",
        "type": "security",
        "config": {
            "schema_type": "event_redflag",
            "sub_schema": "remote_access",
            "condition": 'description ~= "*ssh*"',
            "title": "SSH Login Detected",
            "message": "{{ description }}\n\n{{ __origin__ }}",
            "additional_fields": {
                "details": {
                    "Time": "{{ __hr_time__ }}",
                    "Source UID": "{{ muid }}",
                    "Cluster": "{{ __cluster__ }}",
                    "Logged In As": "{{ euser }}",
                    "From IP": "{{ remote_ip }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":triangular_flag_on_post:",
            },
        },
    },
    {
        "display_name": "Guardian Deviation",
        "id": "nc_tmpl:000000009",
        "description": "Send a notification when we see a guardian deviation. (1 hr cooldown per unique deviation)",
        "type": "guardian",
        "config": {
            "schema_type": "event_audit",
            "sub_schema": "guardian_deviation",
            "condition": "",
            "cooldown": {
                "byField": ["policy_uid", "checksum"],
                "forSeconds": 3600,
            },
            "title": "Spyderbat Guardian Deviation Detected",
            "message": 'Record "{{ ref }}" deviated from policy "{{ policy_name }}". To update the policy using Spyctl execute:\n'
            "```spyctl get deviation --policies {{ policy_uid }} {{ id }} -o yaml > deviation.yaml\n"
            "spyctl merge -p {{ policy_uid }} --with-file deviation.yaml```\n"
            "\n{{ __origin__ }}\n",
            "additional_fields": {
                "details": {
                    "Time": "{{ __hr_time__ }}",
                    "Policy": "{{ policy_name }}",
                    "Policy UID": "{{ policy_uid }}",
                    "Policy Mode": "{{ policy_mode }}",
                    "Cluster": "{{ __cluster__ }}",
                    "Hostname": "{{ hostname }}",
                },
                "linkback_text": "View in Spyderbat",
                "linkback_url": "{{ __linkback__ }}",
                "slack_icon": ":warning:",
            },
        },
    },
]



import json
from typing import List, Union, Literal, Optional

import spyctl.schemas_v2 as schemas
from fastapi import APIRouter, BackgroundTasks
from pydantic import BaseModel, Field, Json
from typing_extensions import Annotated

import app.app_lib as app_lib
import app.commands.diff as cmd_diff

router = APIRouter(prefix="/api/v1")

PrimaryObject = Annotated[
    Union[schemas.GuardianBaselineModel, schemas.GuardianPolicyModel],
    Field(discriminator="kind"),
]

DiffObject = Annotated[
    Union[
        schemas.GuardianBaselineModel,
        schemas.GuardianDeviationModel,
        schemas.GuardianFingerprintGroupModel,
        schemas.GuardianFingerprintModel,
        schemas.GuardianPolicyModel,
        schemas.UidListModel,
    ],
    Field(discriminator="kind"),
]

# ------------------------------------------------------------------------------
# Diff Object(s) into Object
# ------------------------------------------------------------------------------


class DiffHandlerInput(BaseModel):
    object: Json[PrimaryObject] = Field(title="The primary object of the diff")
    diff_objects: Json[List[DiffObject]] = Field(
        title="The object(s) to diff with the primary object."
    )
    org_uid: str
    api_key: str
    api_url: str
    full_diff: Optional[bool] = False
    content_type: Optional[Literal["text", "json"]] = Field(
        default="string", title="The content type of the diff data"
    )


class DiffHandlerOutput(BaseModel):
    diff_data: str


@router.post("/diff")
def diff(
    i: DiffHandlerInput, background_tasks: BackgroundTasks
) -> DiffHandlerOutput:
    background_tasks.add_task(app_lib.flush_spyctl_log_messages)
    diff_objects = [
        json.loads(obj.json(by_alias=True, exclude_unset=True))
        for obj in i.diff_objects
    ]
    cmd_input = cmd_diff.DiffInput(
        json.loads(i.object.json(by_alias=True, exclude_unset=True)),
        diff_objects,
        i.org_uid,
        i.api_key,
        i.api_url,
        i.full_diff,
        i.content_type,
    )
    output = cmd_diff.diff(cmd_input)
    return DiffHandlerOutput(diff_data=output.diff_data)



from typing import Dict, List
import spyctl.spyctl_lib as lib
import spyctl.api as api
import spyctl.resources.api_filters as _af
import spyctl.config.configs as cfg


def handle_input_data(data: Dict, ctx: cfg.Context = None) -> List[Dict]:
    obj_kind = data.get(lib.KIND_FIELD)
    schema = data.get(lib.SCHEMA_FIELD)
    rv = []
    if obj_kind == lib.POL_KIND:
        rv.append(data)
    elif obj_kind == lib.BASELINE_KIND:
        rv.append(data)
    elif obj_kind == lib.FPRINT_KIND:
        rv.append(data)
    elif obj_kind == lib.DEVIATION_KIND or (
        schema
        and schema.startswith(
            (
                f"{lib.EVENT_AUDIT_PREFIX}:"
                f"{lib.EVENT_AUDIT_SUBTYPE_MAP['deviation']}"
            )
        )
    ):
        if obj_kind is None:
            rv.append(data["deviation"])
        else:
            rv.append(data)
    elif obj_kind == lib.FPRINT_GROUP_KIND:
        rv.extend(__handle_fprint_group_input(data))
    elif obj_kind == lib.UID_LIST_KIND:
        rv.extend(__handle_uid_list_input(data, ctx))
    elif lib.ITEMS_FIELD in data:
        rv.extend(__handle_spyctl_items_input(data))
    return rv


def __handle_fprint_group_input(data: Dict):
    return data[lib.DATA_FIELD][lib.FPRINT_GRP_FINGERPRINTS_FIELD]


def __handle_uid_list_input(data: Dict, ctx: cfg.Context = None):
    if not ctx:
        ctx = cfg.get_current_context()
    pipeline = _af.UID_List.generate_pipeline(data)
    time = (
        data[lib.METADATA_FIELD][lib.METADATA_START_TIME_FIELD],
        data[lib.METADATA_FIELD][lib.METADATA_END_TIME_FIELD],
    )
    src = ctx.global_source
    fprints = list(
        api.get_fingerprints(
            *ctx.get_api_data(), [src], time, pipeline=pipeline
        )
    )
    return fprints


def __handle_spyctl_items_input(data: Dict):
    rv = []
    for item in data[lib.ITEMS_FIELD]:
        rv.extend(handle_input_data(item))
    return rv



from typing import Dict, Optional, List, Union

import spyctl.cli as cli
import spyctl.merge_lib as m_lib
import spyctl.resources.fingerprints as spyctl_fprints
import spyctl.resources.policies as spyctl_policies
import spyctl.spyctl_lib as lib
import spyctl.config.configs as cfg
import spyctl.resources.resources_lib as r_lib
import spyctl.schemas_v2 as schemas

FPRINT_KIND = spyctl_fprints.FPRINT_KIND
GROUP_KIND = spyctl_fprints.GROUP_KIND
BASELINE_KIND = lib.BASELINE_KIND
POLICY_KIND = lib.POL_KIND
BASELINE_META_MERGE_SCHEMA = m_lib.MergeSchema(
    lib.METADATA_FIELD,
    merge_functions={
        lib.METADATA_NAME_FIELD: m_lib.keep_base_value_merge,
        lib.METADATA_TYPE_FIELD: m_lib.all_eq_merge,
        lib.LATEST_TIMESTAMP_FIELD: m_lib.greatest_value_merge,
    },
)
BASELINE_MERGE_SCHEMAS = [BASELINE_META_MERGE_SCHEMA, m_lib.SPEC_MERGE_SCHEMA]


class InvalidBaselineError(Exception):
    pass


class Baseline:
    required_keys = {
        lib.API_FIELD,
        lib.KIND_FIELD,
        lib.METADATA_FIELD,
        lib.SPEC_FIELD,
    }

    def __init__(
        self,
        obj: Dict,
        name: str = None,
        disable_procs=None,
        disable_conns=None,
    ) -> None:
        self.baseline = {}
        for key in self.required_keys:
            if key not in obj:
                raise InvalidBaselineError(f"Missing {key} for input object")
        self.metadata = obj[lib.METADATA_FIELD]
        if name:
            self.metadata[lib.METADATA_NAME_FIELD] = name
        self.spec: Dict = obj[lib.SPEC_FIELD]
        self.__parse_disable_procs(disable_procs)
        self.__parse_disable_conns(disable_conns)

    def spec_dict(self):
        spec_field_names = [
            lib.PROC_POLICY_FIELD,
            lib.NET_POLICY_FIELD,
        ]
        selectors = {}
        other_fields = {}
        pol_fields = {}
        for k, v in self.spec.items():
            if "Selector" in k:
                selectors[k] = v
            if k in spec_field_names:
                continue
            elif k == lib.RESPONSE_FIELD:
                continue
            else:
                other_fields[k] = v
        for name in spec_field_names:
            pol_fields[name] = self.spec[name]
        rv = {}
        rv.update(selectors)
        rv.update(other_fields)
        rv.update(pol_fields)
        return rv

    def as_dict(self) -> Dict:
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: BASELINE_KIND,
            lib.METADATA_FIELD: self.metadata,
            lib.SPEC_FIELD: self.spec_dict(),
        }
        return rv

    def __parse_disable_procs(self, disable_procs: Optional[str]):
        if disable_procs == lib.DISABLE_PROCS_ALL:
            self.spec[lib.DISABLE_PROCS_FIELD] = lib.DISABLE_PROCS_ALL

    def __parse_disable_conns(self, disable_conns: Optional[str]):
        if disable_conns == lib.DISABLE_CONNS_ALL:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_EGRESS:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_INGRESS:
            self.spec[lib.DISABLE_CONNS_FIELD] = lib.INGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE_E:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PRIVATE_I:
            self.spec[lib.DISABLE_PR_CONNS_FIELD] = lib.INGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.DISABLE_CONNS_ALL
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC_E:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.EGRESS_FIELD
        elif disable_conns == lib.DISABLE_CONNS_PUBLIC_I:
            self.spec[lib.DISABLE_PU_CONNS_FIELD] = lib.INGRESS_FIELD


def create_baseline(
    input_data: Union[Dict, List[Dict]],
    name: str = None,
    ctx: cfg.Context = None,
    disable_procs: str = None,
    disable_conns: str = None,
):
    input_objs = []
    if isinstance(input_data, list):
        if len(input_data) == 0:
            cli.err_exit("Nothing to build baseline with")
        for datum in input_data:
            input_objs.extend(r_lib.handle_input_data(datum, ctx))
    else:
        input_objs.extend(r_lib.handle_input_data(input_data, ctx))
    if len(input_objs) == 0:
        cli.err_exit("Nothing to build baseline with")
    merge_object = m_lib.MergeObject(
        input_objs[0],
        BASELINE_MERGE_SCHEMAS,
        None,
        disable_procs=disable_procs,
        disable_conns=disable_conns,
    )
    if len(input_objs) == 1:
        merge_object.asymmetric_merge({})
    else:
        for obj in input_objs[1:]:
            merge_object.symmetric_merge(obj)
    try:
        baseline = Baseline(
            merge_object.get_obj_data(), name, disable_procs, disable_conns
        )
    except InvalidBaselineError as e:
        cli.err_exit(f"Unable to create baseline. {' '.join(e.args)}")
    # Validate the Baseline
    rv = baseline.as_dict()
    if not schemas.valid_object(rv):
        cli.err_exit("Created policy failed validation.")
    return baseline.as_dict()


def merge_baseline(
    baseline: Dict, with_obj: Dict, fingerprints
) -> Optional[m_lib.MergeObject]:
    try:
        _ = Baseline(baseline)
    except InvalidBaselineError as e:
        cli.err_exit(f"Invalid baseline as input. {' '.join(e.args)}")
    with_obj_kind = (
        with_obj.get(lib.KIND_FIELD) if isinstance(with_obj, dict) else None
    )
    base_merge_obj = m_lib.MergeObject(
        baseline, BASELINE_MERGE_SCHEMAS, Baseline
    )
    if with_obj_kind == GROUP_KIND:
        fingerprints = with_obj.get(lib.DATA_FIELD, {}).get(
            spyctl_fprints.FINGERPRINTS_FIELD, []
        )
        for fprint in fingerprints:
            base_merge_obj.asymmetric_merge(fprint)
        if not base_merge_obj.is_valid:
            cli.try_log("Merge was unable to create a valid baseline")
    elif with_obj_kind == BASELINE_KIND:
        try:
            _ = Baseline(with_obj)
        except (
            InvalidBaselineError,
            spyctl_fprints.InvalidFingerprintError,
        ) as e:
            cli.err_exit(
                "Invalid baseline object as 'with object' input."
                f" {' '.join(e.args)}"
            )
        base_merge_obj.asymmetric_merge(with_obj)
        if not base_merge_obj.is_valid:
            cli.try_log("Merge was unable to create a valid baseline")
    elif with_obj == POLICY_KIND:
        try:
            _ = spyctl_policies.Policy(with_obj)
        except spyctl_policies.InvalidPolicyError as e:
            cli.err_exit(
                "Invalid policy object as 'with object' input."
                f" {' '.join(e.args)}"
            )
        base_merge_obj.asymmetric_merge(with_obj)
        if not base_merge_obj.is_valid:
            cli.try_log("Merge was unable to create a valid baseline")
    elif fingerprints is not None:
        for fingerprint in fingerprints:
            base_merge_obj.asymmetric_merge(fingerprint)
        if not base_merge_obj.is_valid:
            cli.try_log("Merge was unable to create a valid baseline")
    else:
        cli.try_log(
            f"Merging baseline with {with_obj_kind} is not yet supported."
        )
        return
    return base_merge_obj


def diff_baseline(baseline: Dict, with_obj: Dict, latest):
    base_merge_obj = merge_baseline(baseline, with_obj, latest)
    if not base_merge_obj:
        cli.err_exit("Unable to perform Diff")
    diff = base_merge_obj.get_diff()
    return diff



# A hidden command, used by support to assist with migrating documents
# that need to be updated

from typing import Optional
import spyctl.api as api
import spyctl.config.configs as cfg
import yaml
from pathlib import Path
import spyctl.resources.policies as p
import spyctl.spyctl_lib as lib
import spyctl.commands.apply as apply


def handle_update_response_actions(backup_dir: Optional[str]):
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    if backup_dir:
        backup_path = Path(backup_dir)

    for policy in policies:
        pol_data = p.policies_output([policy])
        if backup_path:
            try:
                uid = pol_data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(pol_data, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        del pol_data[lib.SPEC_FIELD][lib.RESPONSE_FIELD]
        pol_data = p.Policy(pol_data).as_dict()
        apply.handle_apply_policy(pol_data)


def handle_update_policy_modes(backup_dir: Optional[str]):
    ctx = cfg.get_current_context()
    print("Updating Guardian Policies")
    policies = api.get_policies(*ctx.get_api_data())
    if backup_dir:
        backup_path = Path(backup_dir)
    for policy in policies:
        pol_data = p.policies_output([policy])
        if backup_path:
            try:
                uid = pol_data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(pol_data, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        mode = pol_data[lib.SPEC_FIELD].get(
            lib.POL_MODE_FIELD, lib.POL_MODE_ENFORCE
        )
        if mode not in lib.POL_MODES:
            mode = lib.POL_MODE_ENFORCE
        lib.try_log(
            f"Setting mode for policy"
            f" '{policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]}' to"
            f" '{mode}'"
        )
        pol_data[lib.SPEC_FIELD][lib.POL_MODE_FIELD] = mode
        pol_data = p.Policy(pol_data).as_dict()
        apply.handle_apply_policy(pol_data)
    print("Updating Suppression Policies")
    policies = api.get_policies(
        *ctx.get_api_data(),
        params={lib.METADATA_TYPE_FIELD: lib.POL_TYPE_TRACE},
    )
    for policy in policies:
        if backup_path:
            try:
                uid = policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(policy, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        mode = policy[lib.SPEC_FIELD].get(
            lib.POL_MODE_FIELD, lib.POL_MODE_ENFORCE
        )
        if mode not in lib.POL_MODES:
            mode = lib.POL_MODE_ENFORCE
        lib.try_log(
            f"Setting mode for suppression policy"
            f" '{policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]}' to"
            f" '{mode}'"
        )
        policy[lib.SPEC_FIELD][lib.POL_MODE_FIELD] = mode
        apply.handle_apply_suppression_policy(policy)



import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib
import spyctl.filter_resource as filt
import spyctl.resources.notification_targets as nt
from typing import Dict, List

INTERACTIVE_SUPPORTED = [
    lib.NOTIFICATION_TARGETS_RESOURCE,
    lib.NOTIFICATION_CONFIGS_RESOURCE,
]


def handle_delete(resource, name_or_id):
    if resource == lib.NOTIFICATION_CONFIGS_RESOURCE:
        handle_delete_notif_config(name_or_id)
    elif resource == lib.NOTIFICATION_TARGETS_RESOURCE:
        handle_delete_notif_tgt(name_or_id)
    elif resource == lib.POLICIES_RESOURCE:
        handle_delete_policy(name_or_id)
    elif resource == lib.SUPPRESSION_POLICY_RESOURCE:
        handle_delete_suppression_policy(name_or_id)
    else:
        cli.err_exit(f"The 'delete' command is not supported for '{resource}'")


def handle_delete_notif_config(name_or_id):
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    routes: List = notif_pol.get(lib.ROUTES_FIELD, [])
    del_index = None
    del_id = None
    for i, route in enumerate(routes):
        id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        name = route.get(lib.DATA_FIELD, {}).get(lib.NAME_FIELD)
        if id == name_or_id or name == name_or_id:
            if del_index is not None and name == name_or_id:
                cli.err_exit(f"{name_or_id} is ambiguous, use ID")
            del_index = i
            del_id = id
    if del_index is None:
        cli.err_exit(f"No notification config matching '{name_or_id}'")
    if cli.query_yes_no(
        f"Are you sure you want to delete notification config {del_id}"
    ):
        routes.pop(del_index)
        notif_pol[lib.ROUTES_FIELD] = routes
        api.put_notification_policy(*ctx.get_api_data(), notif_pol)
        cli.try_log(f"Successfully deleted notification config '{del_id}'")


def handle_delete_notif_tgt(name_or_id):
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    targets: Dict = notif_pol.get(lib.TARGETS_FIELD, {})
    del_name = None
    # check if name exists
    if name_or_id in targets:
        del_name = name_or_id
    if not del_name:
        for tgt_name, tgt in targets.items():
            id = tgt.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
            if id is None:
                continue
            if id == name_or_id:
                del_name = tgt_name
    if not del_name:
        cli.err_exit(f"No notification target matching '{name_or_id}'.")
    if cli.query_yes_no(
        "Are you sure you want to delete notification target" f" '{del_name}'?"
    ):
        notif_pol = api.get_notification_policy(*ctx.get_api_data())
        notif_pol[lib.TARGETS_FIELD].pop(del_name)
        api.put_notification_policy(*ctx.get_api_data(), notif_pol)
        cli.try_log(f"Successfully deleted notification target '{del_name}'")


def handle_delete_policy(name_or_uid):
    ctx = cfg.get_current_context()

    policies = api.get_policies(*ctx.get_api_data())
    policies = [
        (
            p[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
            p[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
        )
        for p in filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_uid,
        )
    ]
    if len(policies) == 0:
        cli.err_exit(f"No policies matching name_or_uid '{name_or_uid}'")
    for uid, name in policies:
        perform_delete = cli.query_yes_no(
            f"Are you sure you want to delete policy '{name} - {uid}' from"
            " Spyderbat?"
        )
        if perform_delete:
            api.delete_policy(
                *ctx.get_api_data(),
                uid,
            )
            cli.try_log(f"Successfully deleted policy '{name} - {uid}'")
        else:
            cli.try_log(f"Skipping delete of '{name} - {uid}'")


def handle_delete_suppression_policy(name_or_uid):
    ctx = cfg.get_current_context()

    policies = api.get_policies(
        *ctx.get_api_data(), {lib.TYPE_FIELD: lib.POL_TYPE_TRACE}
    )
    policies = [
        (
            p[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
            p[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
        )
        for p in filt.filter_obj(
            policies,
            [
                [lib.METADATA_FIELD, lib.METADATA_NAME_FIELD],
                [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
            ],
            name_or_uid,
        )
    ]
    if len(policies) == 0:
        cli.err_exit(
            f"No suppression policies matching name_or_uid '{name_or_uid}'"
        )
    for uid, name in policies:
        perform_delete = cli.query_yes_no(
            f"Are you sure you want to delete suppression policy "
            f"'{name} - {uid}' from Spyderbat?"
        )
        if perform_delete:
            api.delete_policy(
                *ctx.get_api_data(),
                uid,
            )
            cli.try_log(f"Successfully deleted policy '{name} - {uid}'")
        else:
            cli.try_log(f"Skipping delete of '{name} - {uid}'")



import json
from typing import List, Literal, Optional, Union

import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib
from fastapi import APIRouter, BackgroundTasks
from pydantic import BaseModel, Field, Json
from typing_extensions import Annotated
import app.app_lib as app_lib

import app.commands.create as cmd_create

router = APIRouter(prefix="/api/v1")

# ------------------------------------------------------------------------------
# Create Suppression Policy
# ------------------------------------------------------------------------------


def undash(s: str) -> str:
    return s.replace("-", "_")


class CreateSuppressionPolicyInputSelectorFields(BaseModel):
    trigger_ancestors: Optional[List[str]] = Field(
        alias=undash(lib.SUP_POL_CMD_TRIG_ANCESTORS)
    )
    trigger_class: Optional[List[str]] = Field(
        alias=undash(lib.SUP_POL_CMD_TRIG_CLASS)
    )
    users: Optional[List[str]] = Field(alias=undash(lib.SUP_POL_CMD_USERS))
    interactive_users: Optional[List[str]] = Field(
        alias=undash(lib.SUP_POL_CMD_INT_USERS)
    )
    non_interactive_users: Optional[List[str]] = Field(
        alias=undash(lib.SUP_POL_CMD_N_INT_USERS)
    )


class CreateSuppressionPolicyHandlerInput(BaseModel):
    type: Literal[tuple(lib.SUPPRESSION_POL_TYPES)] = Field(  # type: ignore
        title="The type of suppression policy to create"
    )
    object_uid: str | None = Field(title="UID of the object to suppress")
    scope_to_users: bool = Field(
        default=False, title="Scope the created policy to the relevant users"
    )
    selectors: Optional[CreateSuppressionPolicyInputSelectorFields] = Field(
        default={}, title="Additional selectors to add to the policy"
    )
    name: str = Field(
        default="", title="Optional name for the suppression policy"
    )
    org_uid: str
    api_key: str
    api_url: str


class CreateSuppressionPolicyHandlerOutput(BaseModel):
    policy: str = Field(default="", title="The policy that was created")


@router.post("/create/suppressionpolicy")
def create_suppression_policy(
    i: CreateSuppressionPolicyHandlerInput, background_tasks: BackgroundTasks
) -> CreateSuppressionPolicyHandlerOutput:
    background_tasks.add_task(app_lib.flush_spyctl_log_messages)
    cmd_input = cmd_create.CreateSuppressionPolicyInput(
        i.type,
        i.object_uid,
        i.scope_to_users,
        i.name,
        i.org_uid,
        i.api_key,
        i.api_url,
        i.selectors.dict(by_alias=True, exclude_unset=True),
    )
    output = cmd_create.suppression_policy(cmd_input)
    return CreateSuppressionPolicyHandlerOutput(policy=output.policy)


# ------------------------------------------------------------------------------
# Create Guardian Policy
# ------------------------------------------------------------------------------

InputObject = Annotated[
    Union[
        schemas.GuardianBaselineModel,
        schemas.GuardianFingerprintModel,
        schemas.GuardianFingerprintGroupModel,
        schemas.UidListModel,
    ],
    Field(discriminator="kind"),
]


class CreateGuardianPolicyHandlerInput(BaseModel):
    input_objects: Json[List[InputObject]] = Field(
        title="The input object(s) used to build the policy"
    )
    name: str = Field(
        default="", title="Optional name for the guardian policy"
    )
    mode: Literal[tuple(lib.POL_MODES)] = Field(  # type: ignore
        default=lib.POL_MODE_AUDIT,
        title="Determines whether a policy is in enforce or audit mode.",
    )
    org_uid: str
    api_key: str
    api_url: str


class CreateGuardianPolicyHandlerOutput(BaseModel):
    policy: str = Field(default="", title="The policy that was created")


@router.post("/create/guardianpolicy")
def create_guardian_policy(
    i: CreateGuardianPolicyHandlerInput, background_tasks: BackgroundTasks
) -> CreateGuardianPolicyHandlerOutput:
    background_tasks.add_task(app_lib.flush_spyctl_log_messages)
    input_objects = [
        json.loads(obj.json(by_alias=True, exclude_unset=True))
        for obj in i.input_objects
    ]
    cmd_input = cmd_create.CreateGuardianPolicyInput(
        i.name,
        input_objects,
        i.mode,
        i.org_uid,
        i.api_key,
        i.api_url,
    )
    output = cmd_create.guardian_policy(cmd_input)
    return CreateGuardianPolicyHandlerOutput(policy=output.policy)



import json
from dataclasses import dataclass, field

import spyctl.commands.create as spyctl_create
import spyctl.spyctl_lib as lib
from fastapi import HTTPException
from typing import List, Dict

import app.app_lib as app_lib

# ------------------------------------------------------------------------------
# Create Suppression Policy
# ------------------------------------------------------------------------------


@dataclass
class CreateSuppressionPolicyInput:
    type: str
    obj_uid: str = ""
    auto_generate_user_scope: bool = False
    name: str = ""
    org_uid: str = ""
    api_key: str = ""
    api_url: str = ""
    selectors: dict = field(default_factory=dict)


@dataclass
class CreateSuppressionPolicyOutput:
    policy: str


def suppression_policy(
    i: CreateSuppressionPolicyInput,
) -> CreateSuppressionPolicyOutput:
    if i.type == lib.POL_TYPE_TRACE:
        return trace_suppression_policy(i)
    else:
        raise HTTPException(
            status_code=400,
            detail="Bad Request. Invalid suppression policy type.",
        )


def trace_suppression_policy(
    i: CreateSuppressionPolicyInput,
) -> CreateSuppressionPolicyOutput:
    spyctl_ctx = app_lib.generate_spyctl_context(
        i.org_uid, i.api_key, i.api_url
    )
    pol = spyctl_create.create_trace_suppression_policy(
        i.obj_uid,
        i.auto_generate_user_scope,
        i.name,
        ctx=spyctl_ctx,
        **i.selectors,
    )
    output = CreateSuppressionPolicyOutput(json.dumps(pol.as_dict()))
    return output


# ------------------------------------------------------------------------------
# Create Guardian Policy
# ------------------------------------------------------------------------------


@dataclass
class CreateGuardianPolicyInput:
    name: str = ""
    input_objs: List[Dict] = field(default_factory=list())
    mode: str = lib.POL_MODE_AUDIT
    org_uid: str = ""
    api_key: str = ""
    api_url: str = ""


@dataclass
class CreateGuardianPolicyOutput:
    policy: str


def guardian_policy(
    i: CreateGuardianPolicyInput,
) -> CreateGuardianPolicyOutput:
    spyctl_ctx = app_lib.generate_spyctl_context(
        i.org_uid, i.api_key, i.api_url
    )
    pol = spyctl_create.create_guardian_policy_from_json(
        i.name, i.mode, i.input_objs, spyctl_ctx
    )
    output = CreateGuardianPolicyOutput(policy=json.dumps(pol))
    return output



import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.cli as cli
from typing import Optional, Dict
import spyctl.spyctl_lib as lib
import json


def search_for_trace_by_uid(
    uid: str, ctx: cfg.Context = None
) -> Optional[Dict]:
    if not uid.startswith("trace:"):
        cli.err_exit("Expected uid in search for trace to start with 'trace:'")
    cli.try_log("Searching for trace object...")
    schema = lib.MODEL_SPYDERTRACE_PREFIX
    datatype = lib.DATATYPE_SPYDERGRAPH
    return search_for_obj_by_uid(uid, schema, datatype, ctx)


def search_for_trace_summary_by_uid(
    uid: str, ctx: cfg.Context = None
) -> Optional[Dict]:
    if not uid.startswith("fprint:trace"):
        cli.err_exit(
            "Expected uid in search for trace to start with 'fprint:trace'"
        )
    cli.try_log("Searching for trace summary...")
    schema = lib.MODEL_FINGERPRINT_PREFIX
    datatype = lib.DATATYPE_FINGERPRINTS
    return search_for_obj_by_uid(uid, schema, datatype, ctx)


def search_for_obj_by_uid(
    uid: str, schema: str, datatype: str, ctx: cfg.Context = None
) -> Optional[Dict]:
    if not ctx:
        ctx = cfg.get_current_context()
    resp = api.get_object_by_id(
        *ctx.get_api_data(),
        id=uid,
        schema=schema,
        datatype=datatype,
    )
    if resp.status_code != 200:
        cli.try_log(f"{resp.text}", is_warning=True)
        return None
    else:
        lines = resp.text.splitlines()
        if len(lines) < 2:
            cli.try_log(
                f"Unable to find {schema} object with id {uid}",
                is_warning=True,
            )
            return None
        try:
            obj: Dict = json.loads(lines[1])
        except json.JSONDecodeError:
            cli.try_log(
                "Unable to parse json returned by the spyderbat API",
                is_warning=True,
            )
            return None
    return obj



import json
from dataclasses import dataclass
from typing import Dict, List

from spyctl.commands.merge import merge_resource

import app.app_lib as app_lib
import app.exceptions as ex

# ------------------------------------------------------------------------------
# Diff Object with Object(s)
# ------------------------------------------------------------------------------


@dataclass
class DiffInput:
    object: Dict
    diff_objects: List[Dict]
    org_uid: str = ""
    api_key: str = ""
    api_url: str = ""
    full_diff: bool = False
    content_type: str = "text"


@dataclass
class DiffOutput:
    diff_data: str


def diff(i: DiffInput) -> DiffOutput:
    spyctl_ctx = app_lib.generate_spyctl_context(
        i.org_uid, i.api_key, i.api_url
    )
    merge_data = merge_resource(
        i.object,
        "API Diff Request Object",
        i.diff_objects,
        ctx=spyctl_ctx,
    )
    if not merge_data:
        msg = app_lib.flush_spyctl_log_messages()
        ex.internal_server_error(msg)
    app_lib.flush_spyctl_log_messages()
    if i.content_type == "json":
        diff_obj = True
    else:
        diff_obj = False
    diff_data = merge_data.get_diff(i.full_diff, diff_obj)
    if isinstance(diff_data, str):
        if i.content_type == "json":
            raise ValueError(
                "Diff of this object type does not support JSON output."
            )
        return DiffOutput(diff_data)
    return DiffOutput(json.dumps(diff_data))



import json


from tabulate import tabulate
import zulu


# helper function
def epoch_to_zulu(epoch):
    try:
        return zulu.Zulu.fromtimestamp(epoch).format("YYYY-MM-ddTHH:mm:ss") + "Z"
    except Exception:
        return epoch


# step 1: spyctl get connections -o json --ndjson --remote-port 8126 > connections8126.json and
# spyctl get containers -o json --ndjson > containers8126.json

# load connection data and add it to the conn dictionary
# if any field is not present default is " "
conn_dict = {}
with open("connections8126.json", "r") as file:
    for connections in file:
        connection = json.loads(connections)
        containerUID = connection.get("container_uid", " ")
        cgroup = connection.get("cgroup", " ")
        payload = connection.get("payload", " ")
        conn_dict[containerUID] = {"cgroup": cgroup}
        connectionID = connection.get("id", " ")


# load container data and add it to the container dict with
# containerUID as the key
container_dict = {}
with open("containers8126.json", "r") as file:
    for containers in file:
        containerInfo = json.loads(containers)
        containerUID = containerInfo.get("id", " ")
        containerName = containerInfo.get("container_name", " ")
        container_created = containerInfo.get("created", " ")
        container_created = epoch_to_zulu(container_created)
        cluster = containerInfo.get("clustername", " ")
        pod = containerInfo.get("pod_name", " ")
        image = containerInfo.get("image_runtime", " ")
        state = containerInfo.get("container_state_runtime", " ")
        container_dict[containerUID] = {
            "containerUID": containerUID,
            "containerName": containerName,
            "created": epoch_to_zulu(container_created),
            "state": state,
            "cluster": cluster,
            "pod": pod,
            "image": image,
        }


# match the container--> connection data based on connection key.
matched_data = []

for connection_key, connection_data in conn_dict.items():
    if connection_key in container_dict:
        matched_data.append(
            {
                "containerID": connection_key,
                "container": containerName,
                "containerCreated": container_created,
                "state": state,
                "pod": pod,
                "cluster": cluster,
                "image": image,
                "cgroup": connection_data["cgroup"],
            }
        )


# print the data in tabulate manner.
headers = [
    "containerID",
    "containerName",
    "containerCreated",
    "state",
    "pod",
    "cluster",
    "image",
    "cgroup",
]
table = tabulate(
    [list(d.values()) for d in matched_data], headers=headers, tablefmt="plain"
)

print(table)



from typing import IO, Dict, List, Optional, Union

import spyctl.api as api
import spyctl.cli as cli
import spyctl.commands.apply as apply
import spyctl.commands.get as get
import spyctl.config.configs as cfgs
import spyctl.filter_resource as filt
import spyctl.merge_lib as m_lib
import spyctl.resources.baselines as b
import spyctl.resources.policies as p
import spyctl.resources.suppression_policies as sp
import spyctl.resources.deviations as dev
import spyctl.resources.api_filters as _af
import spyctl.spyctl_lib as lib
import spyctl.schemas_v2 as schemas
import spyctl.resources.resources_lib as r_lib

POLICIES = None
FINGERPRINTS = None
DEVIATIONS = None
MATCHING = "matching"
ALL = "all"

YES_EXCEPT = False


def handle_merge(
    filename_target: List[IO],
    policy_target: List[str],
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest: bool,
    output: str,
    output_to_file: bool = False,
    yes_except: bool = False,
    merge_network: bool = True,
    do_api=False,
    force_fprints=False,
    full_diff=False,
):
    if do_api:
        ctx = cfgs.get_current_context()
        if not filename_target or not with_file:
            cli.err_exit("api test only with local files")
        r_data = lib.load_file_for_api_test(filename_target[0])
        w_data = lib.load_file_for_api_test(with_file)
        merged_data = api.api_merge(*ctx.get_api_data(), r_data, w_data)
        cli.show(merged_data, lib.OUTPUT_RAW)
        return
    global POLICIES, YES_EXCEPT
    YES_EXCEPT = yes_except
    if not POLICIES and (with_policy or policy_target):
        ctx = cfgs.get_current_context()
        POLICIES = api.get_policies(*ctx.get_api_data())
        POLICIES.sort(key=lambda x: x[lib.METADATA_FIELD][lib.NAME_FIELD])
        if not POLICIES and policy_target:
            cli.err_exit("No policies to merge.")
        elif not POLICIES and with_policy:
            cli.err_exit("No policies to merge with.")
    if filename_target:
        if len(filename_target) > 1 or output_to_file:
            output_dest = lib.OUTPUT_DEST_FILE
        else:
            output_dest = lib.OUTPUT_DEFAULT
        filename_target.sort(key=lambda x: x.name)
        for file in filename_target:
            target = load_target_file(file)
            target_name = f"local file '{file.name}'"
            resrc_kind = target.get(lib.KIND_FIELD)
            if resrc_kind not in [lib.BASELINE_KIND, lib.POL_KIND]:
                cli.try_log(
                    f"The 'merge' command is not supported for {resrc_kind}",
                    is_warning=True,
                )
                continue
            with_obj = get_with_obj(
                target,
                target_name,
                with_file,
                with_policy,
                st,
                et,
                latest,
                with_fingerprints=force_fprints,
            )
            # If we have something to merge, add to actions
            if with_obj:
                merged_obj = merge_resource(
                    target, target_name, with_obj, merge_network=merge_network
                )
                if merged_obj:
                    handle_output(
                        output, output_dest, merged_obj, full_diff=full_diff
                    )
            elif with_obj is False:
                continue
            else:
                merge_obj = __nothing_to_merge_with(
                    target_name, target, latest
                )
                if merge_obj:
                    handle_output(
                        output, output_dest, merge_obj, full_diff=full_diff
                    )
    elif policy_target:
        if output_to_file:
            output_dest = lib.OUTPUT_DEST_FILE
        else:
            output_dest = lib.OUTPUT_DEST_API
        policy_target = sorted(policy_target)
        if ALL in policy_target:
            for target in POLICIES:
                t_name = lib.get_metadata_name(target)
                t_uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    output_dest,
                    with_fingerprints=force_fprints,
                )
                # If we have something to merge, add to actions
                if with_obj:
                    merged_obj = merge_resource(
                        target,
                        target_name,
                        with_obj,
                        merge_network=merge_network,
                    )
                    if merged_obj:
                        handle_output(
                            output,
                            output_dest,
                            merged_obj,
                            full_diff=full_diff,
                        )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_merge_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(
                            output, output_dest, merge_obj, full_diff=full_diff
                        )
        else:
            targets = {}
            for pol_name_or_uid in policy_target:
                policies = filt.filter_obj(
                    POLICIES,
                    [
                        [lib.METADATA_FIELD, lib.NAME_FIELD],
                        [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                    ],
                    pol_name_or_uid,
                )
                if len(policies) == 0:
                    cli.try_log(
                        "Unable to locate policy with name or UID"
                        f" {pol_name_or_uid}",
                        is_warning=True,
                    )
                    continue
                for policy in policies:
                    pol_uid = policy[lib.METADATA_FIELD][
                        lib.METADATA_UID_FIELD
                    ]
                    targets[pol_uid] = policy
            targets = sorted(
                list(targets.values()),
                key=lambda x: x[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
            )
            if len(targets) > 0:
                pager = True
            for target in targets:
                t_name = lib.get_metadata_name(target)
                t_uid = target.get(lib.METADATA_FIELD, {}).get(
                    lib.METADATA_UID_FIELD
                )
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    output_dest,
                    with_fingerprints=force_fprints,
                )
                if with_obj:
                    merged_obj = merge_resource(
                        target,
                        target_name,
                        with_obj,
                        merge_network=merge_network,
                    )
                    if merged_obj:
                        handle_output(
                            output,
                            output_dest,
                            merged_obj,
                            pager,
                            full_diff=full_diff,
                        )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_merge_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(
                            output,
                            output_dest,
                            merge_obj,
                            pager,
                            full_diff=full_diff,
                        )
    else:
        cli.err_exit("No target(s) to merge.")


def get_with_obj(
    target: Dict,
    target_name: str,
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
    dest: str = "",
    with_fingerprints=False,
) -> Optional[Union[Dict, List[Dict], bool]]:
    global FINGERPRINTS
    target_uid = target.get(lib.METADATA_FIELD, {}).get(lib.METADATA_UID_FIELD)
    if dest == lib.OUTPUT_DEST_API:
        apply_disclaimer = (
            " (Note: you will have a chance to review any merge changes before"
            " applying them.)"
        )
    else:
        apply_disclaimer = ""
    if with_file:
        with_obj = load_with_file(with_file)
        if not cli.query_yes_no(
            f"Merge {target_name} with local file '{with_file.name}'?"
            f"{apply_disclaimer}"
        ):
            return False
    elif with_policy == MATCHING:
        if not target_uid:
            cli.try_log(
                f"{target_name} has no uid, unable to match with"
                " policy stored in the Spyderbat backend.",
                is_warning=True,
            )
            return False
        with_obj = get_with_policy(target_uid, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"Merge {target_name} with data from applied policy "
                f"'{pol_name} - {pol_uid}'?{apply_disclaimer}"
            ):
                return False
    elif with_policy:
        with_obj = get_with_policy(with_policy, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"Merge {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?{apply_disclaimer}"
            ):
                return False
    elif with_fingerprints or not target.get(lib.METADATA_FIELD, {}).get(
        lib.METADATA_UID_FIELD
    ):
        if latest:
            st = get_latest_timestamp(target)
            if FINGERPRINTS is not None:
                cli.try_log("--latest flag set, re-downloading fingerprints..")
                FINGERPRINTS = None
        if not cli.query_yes_no(
            f"Merge {target_name} with relevant Fingerprints from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
            f"{apply_disclaimer}"
        ):
            return False
        if FINGERPRINTS is None:
            with_obj = get_with_fingerprints(target, st, et, latest)
            cli.try_log(f"Filtering fingerprints for {target_name}")
            with_obj = filter_fingerprints(target, with_obj)
        else:
            cli.try_log(f"Filtering fingerprints for {target_name}")
            with_obj = filter_fingerprints(target, FINGERPRINTS)
    else:
        if latest:
            st = get_latest_timestamp(target)
        if not cli.query_yes_no(
            f"Merge {target_name} with Deviations from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
            f"{apply_disclaimer}"
        ):
            return False
        uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
        with_obj = get_with_deviations(uid, st, et)
    return with_obj


def load_target_file(target_file: IO) -> Dict:
    rv = lib.load_resource_file(target_file)
    return rv


def get_target_policy(target_uid) -> Optional[Dict]:
    rv = p.get_policy_by_uid(target_uid, POLICIES)
    return rv


def load_with_file(with_file: IO) -> Dict:
    rv = lib.load_resource_file(with_file)
    return rv


def get_latest_timestamp(target: Dict) -> float:
    latest_timestamp = target.get(lib.METADATA_FIELD, {}).get(
        lib.LATEST_TIMESTAMP_FIELD
    )
    if latest_timestamp is not None:
        st = lib.time_inp(latest_timestamp)
    else:
        cli.err_exit(
            f"No {lib.LATEST_TIMESTAMP_FIELD} found in provided"
            f" resource {lib.METADATA_FIELD} field. Defaulting to"
            " 24hrs."
        )
    return st


def get_with_fingerprints(target: Dict, st, et, latest) -> List[Dict]:
    global FINGERPRINTS
    ctx = cfgs.get_current_context()
    if latest:
        filters = lib.selectors_to_filters(target)
        muids = get.get_muids_scope(**filters)
        if muids:
            filters[lib.MACHINES_FIELD] = muids
        pipeline = _af.Fingerprints.generate_pipeline(filters=filters)
        fingerprints = list(
            api.get_fingerprints(
                *ctx.get_api_data(),
                [ctx.global_source],
                time=(st, et),
                pipeline=pipeline,
            )
        )
    else:
        fingerprints = list(
            api.get_fingerprints(
                *ctx.get_api_data(),
                [ctx.global_source],
                time=(st, et),
            )
        )
    FINGERPRINTS = fingerprints
    return fingerprints


def get_with_deviations(uid: str, st, et) -> List[Dict]:
    deviations = dev.get_unique_deviations(uid, st, et)
    return deviations


def filter_fingerprints(target, fingerprints) -> List[Dict]:
    filters = lib.selectors_to_filters(target)
    rv = filt.filter_fingerprints(
        fingerprints, **filters, use_context_filters=False
    )
    return rv


def get_with_policy(pol_uid: str, policies: List[Dict]) -> Dict:
    return p.get_policy_by_uid(pol_uid, policies)


def merge_resource(
    target: Dict,
    target_name: str,
    with_obj: Union[Dict, List[Dict]],
    src_cmd="merge",
    merge_network=True,
    ctx: cfgs.Context = None,
    latest=False,
) -> Optional[m_lib.MergeObject]:
    if target == with_obj:
        cli.try_log(
            f"{src_cmd} target and with-object are the same.. skipping"
        )
        return None
    if not ctx:
        ctx = cfgs.get_current_context()
    merge_with_objects = []
    if isinstance(with_obj, Dict):
        merge_with_objects = r_lib.handle_input_data(with_obj, ctx)
    else:
        for obj in with_obj:
            merge_with_objects.extend(
                r_lib.handle_input_data(data=obj, ctx=ctx)
            )
    resrc_kind = target.get(lib.KIND_FIELD)
    merge_obj = get_merge_object(resrc_kind, target, merge_network, src_cmd)
    if isinstance(merge_with_objects, list):
        for w_obj in merge_with_objects:
            if is_type_mismatch(target, target_name, src_cmd, w_obj):
                continue
            try:
                merge_obj.asymmetric_merge(w_obj)
            except m_lib.InvalidMergeError as e:
                cli.try_log(
                    f"Unable to {src_cmd} with invalid object. {w_obj}",
                    *e.args,
                )
    else:
        raise Exception(
            f"Bug found, attempting to {src_cmd} with invalid object"
        )
    if target[lib.SPEC_FIELD] == merge_obj.get_obj_data().get(lib.SPEC_FIELD):
        if latest and src_cmd == "merge":
            cli.try_log(
                f"{src_cmd} of {target_name} produced no updates to the"
                f" '{lib.SPEC_FIELD}'.. updating {lib.LATEST_TIMESTAMP_FIELD}"
                " field to now."
            )
            merge_obj.update_latest_timestamp()
            return merge_obj
        else:
            cli.try_log(
                f"{src_cmd} of {target_name} produced no updates to the"
                f" '{lib.SPEC_FIELD}' field.. skipping"
            )
        if lib.API_CALL:
            return merge_obj
        return None
    return merge_obj


def get_merge_object(
    resrc_kind: str, target: Dict, merge_network: bool, src_cmd: str
):
    if resrc_kind == lib.BASELINE_KIND:
        merge_obj = m_lib.MergeObject(
            target,
            b.BASELINE_MERGE_SCHEMAS,
            schemas.valid_object,
            merge_network,
        )
    elif resrc_kind == lib.POL_KIND:
        resrc_type = target[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
        if resrc_type == lib.POL_TYPE_TRACE:
            merge_schemas = sp.T_S_POLICY_MERGE_SCHEMAS
        else:
            merge_schemas = p.POLICY_MERGE_SCHEMAS
        merge_obj = m_lib.MergeObject(
            target, merge_schemas, schemas.valid_object, merge_network
        )
    else:
        cli.try_log(
            f"The '{src_cmd}' command is not supported for {resrc_kind}",
            is_warning=True,
        )
    return merge_obj


def is_type_mismatch(
    target: Dict, target_name: str, src_cmd: str, with_obj: Dict
) -> bool:
    resrc_type = target[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    with_type = with_obj[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    with_kind = with_obj.get(lib.KIND_FIELD)
    target_kind = target.get(lib.KIND_FIELD)
    if target_kind == lib.POL_KIND and with_kind == lib.DEVIATION_KIND:
        pol_uid = target[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
        dev_pol_uid = with_obj[lib.METADATA_FIELD]["policy_uid"]
        if pol_uid != dev_pol_uid:
            cli.try_log(
                f"Error uid mismatch. Trying to {src_cmd} '{target_name}' of"
                f" type '{resrc_type}' with '{with_kind}' but the policy uids"
                " do not match. Skipping...",
                is_warning=True,
            )
            return True
        return False
    if resrc_type != with_type:
        cli.try_log(
            f"Error type mismatch. Trying to {src_cmd} '{target_name}' of type"
            f" '{resrc_type}' with '{with_kind}' object of type '{with_type}'."
            " Skipping...",
            is_warning=True,
        )
        return True
    return False


def handle_output(
    output_format: str,
    output_dest: str,
    merge_obj: m_lib.MergeObject,
    pager=False,
    full_diff=False,
):
    data = merge_obj.get_obj_data()
    if output_dest == lib.OUTPUT_DEST_API:
        apply_merge(merge_obj)
    elif output_dest == lib.OUTPUT_DEST_FILE:
        save_merge_to_file(merge_obj, output_format)
    else:
        if pager:
            cli.show(data, output_format, dest=lib.OUTPUT_DEST_PAGER)
        else:
            cli.show(data, output_format)


def apply_merge(merge_obj: m_lib.MergeObject, full_diff=False):
    data = merge_obj.get_obj_data()
    pol_name = lib.get_metadata_name(data)
    pol_uid = data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
    if (YES_EXCEPT or not cli.YES_OPTION) and cli.query_yes_no(
        f"Review merge updates to '{pol_name}-{pol_uid}'?",
        default="no",
        ignore_yes_option=YES_EXCEPT,
    ):
        cli.show(
            merge_obj.get_diff(full_diff),
            lib.OUTPUT_RAW,
            dest=lib.OUTPUT_DEST_PAGER,
        )
    if not cli.query_yes_no(
        f"Apply merge changes to '{pol_name}-{pol_uid}'?",
        default=None,
        ignore_yes_option=YES_EXCEPT,
    ):
        return
    apply.handle_apply_policy(data)


def save_merge_to_file(merge_obj: m_lib.MergeObject, output_format):
    data = merge_obj.get_obj_data()
    pol_name = lib.get_metadata_name(data)
    pol_uid = data[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    pol_str = f"{pol_name}-{pol_uid}" if pol_uid else pol_name
    if (YES_EXCEPT or not cli.YES_OPTION) and cli.query_yes_no(
        f"Review merge updates to '{pol_str}'?",
        default="no",
        ignore_yes_option=YES_EXCEPT,
    ):
        cli.show(
            merge_obj.get_diff(), lib.OUTPUT_RAW, dest=lib.OUTPUT_DEST_PAGER
        )
    if not cli.query_yes_no(
        f"Apply merge changes to '{pol_name}-{pol_uid}'?",
        default=None,
        ignore_yes_option=YES_EXCEPT,
    ):
        return
    out_fn = lib.find_resource_filename(data, "merge_output")
    out_fn = lib.unique_fn(out_fn, output_format)
    cli.show(data, output_format, dest=lib.OUTPUT_DEST_FILE, output_fn=out_fn)


def __nothing_to_merge_with(
    name: str, target, latest, src_cmd="merge"
) -> Optional[m_lib.MergeObject]:
    if latest:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with. Updating"
            f" '{lib.LATEST_TIMESTAMP_FIELD}' field."
        )
        merge_object = get_merge_object(
            target[lib.KIND_FIELD], target, True, src_cmd
        )
        merge_object.update_latest_timestamp()
        return merge_object
    else:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with... skipping."
        )
